# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
import sys
from unittest.mock import Mock

from azure.core.credentials import AccessToken
from azure.core.pipeline import AsyncPipeline
from azure.core.pipeline.policies import (
    SansIOHTTPPolicy,
    UserAgentPolicy,
    DistributedTracingPolicy,
    AsyncBearerTokenCredentialPolicy,
    AsyncRetryPolicy,
    AsyncRedirectPolicy,
    AsyncHTTPPolicy,
    AsyncRetryPolicy,
    HttpLoggingPolicy,
    SensitiveHeaderCleanupPolicy,
)
from azure.core.pipeline.transport import (
    AsyncHttpTransport,
    AsyncioRequestsTransport,
    TrioRequestsTransport,
    AioHttpTransport,
    HttpRequest,
)

from azure.core.polling.async_base_polling import AsyncLROBasePolling
from azure.core.polling.base_polling import LocationPolling

from azure.core.configuration import Configuration
from azure.core import AsyncPipelineClient
from azure.core.exceptions import AzureError

import aiohttp
import trio

import pytest
from utils import HTTP_REQUESTS


try:
    from unittest.mock import AsyncMock, PropertyMock
except ImportError:
    pass


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_sans_io_exception(http_request):
    class BrokenSender(AsyncHttpTransport):
        async def send(self, request, **config):
            raise ValueError("Broken")

        async def open(self):
            self.session = requests.Session()

        async def close(self):
            self.session.close()

        async def __aexit__(self, exc_type, exc_value, traceback):
            """Raise any exception triggered within the runtime context."""
            return self.close()

    pipeline = AsyncPipeline(BrokenSender(), [SansIOHTTPPolicy()])

    req = http_request("GET", "/")
    with pytest.raises(ValueError):
        await pipeline.run(req)

    class SwapExec(SansIOHTTPPolicy):
        def on_exception(self, requests, **kwargs):
            exc_type, exc_value, exc_traceback = sys.exc_info()
            raise NotImplementedError(exc_value)

    pipeline = AsyncPipeline(BrokenSender(), [SwapExec()])
    with pytest.raises(NotImplementedError):
        await pipeline.run(req)


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_basic_aiohttp(port, http_request):

    request = http_request("GET", "http://localhost:{}/basic/string".format(port))
    policies = [UserAgentPolicy("myusergant"), AsyncRedirectPolicy()]
    async with AsyncPipeline(AioHttpTransport(), policies=policies) as pipeline:
        response = await pipeline.run(request)

    assert pipeline._transport.session is None
    # all we need to check is if we are able to make the call
    assert isinstance(response.http_response.status_code, int)


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_basic_aiohttp_separate_session(port, http_request):

    session = aiohttp.ClientSession()
    request = http_request("GET", "http://localhost:{}/basic/string".format(port))
    policies = [UserAgentPolicy("myusergant"), AsyncRedirectPolicy()]
    transport = AioHttpTransport(session=session, session_owner=False)
    async with AsyncPipeline(transport, policies=policies) as pipeline:
        response = await pipeline.run(request)

    assert transport.session
    assert isinstance(response.http_response.status_code, int)
    await transport.close()
    assert transport.session
    await transport.session.close()


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_basic_async_requests(port, http_request):

    request = http_request("GET", "http://localhost:{}/basic/string".format(port))
    policies = [UserAgentPolicy("myusergant"), AsyncRedirectPolicy()]
    async with AsyncPipeline(AsyncioRequestsTransport(), policies=policies) as pipeline:
        response = await pipeline.run(request)

    assert isinstance(response.http_response.status_code, int)


@pytest.mark.asyncio
async def test_async_transport_sleep():

    async with AsyncioRequestsTransport() as transport:
        await transport.sleep(1)

    async with AioHttpTransport() as transport:
        await transport.sleep(1)


def test_polling_with_path_format_arguments():
    method = AsyncLROBasePolling(timeout=0, path_format_arguments={"host": "host:3000", "accountName": "local"})
    client = AsyncPipelineClient(base_url="http://{accountName}{host}")

    method._operation = LocationPolling()
    method._operation._location_url = "/results/1"
    method._client = client
    assert "http://localhost:3000/results/1" == method._client.format_url(
        method._operation.get_polling_url(), **method._path_format_arguments
    )


def test_async_trio_transport_sleep():
    async def do():
        async with TrioRequestsTransport() as transport:
            await transport.sleep(1)

    response = trio.run(do)


def test_default_http_logging_policy():
    config = Configuration()
    pipeline_client = AsyncPipelineClient(base_url="test")
    pipeline = pipeline_client._build_pipeline(config)
    http_logging_policy = pipeline._impl_policies[-1]._policy
    assert http_logging_policy.allowed_header_names == HttpLoggingPolicy.DEFAULT_HEADERS_WHITELIST
    assert http_logging_policy.allowed_header_names == HttpLoggingPolicy.DEFAULT_HEADERS_ALLOWLIST


def test_pass_in_http_logging_policy():
    config = Configuration()
    http_logging_policy = HttpLoggingPolicy()
    http_logging_policy.allowed_header_names.update({"x-ms-added-header"})
    config.http_logging_policy = http_logging_policy

    pipeline_client = AsyncPipelineClient(base_url="test")
    pipeline = pipeline_client._build_pipeline(config)
    http_logging_policy = pipeline._impl_policies[-1]._policy
    assert http_logging_policy.allowed_header_names == HttpLoggingPolicy.DEFAULT_HEADERS_WHITELIST.union(
        {"x-ms-added-header"}
    )
    assert http_logging_policy.allowed_header_names == HttpLoggingPolicy.DEFAULT_HEADERS_ALLOWLIST.union(
        {"x-ms-added-header"}
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_conf_async_requests(port, http_request):

    request = http_request("GET", "http://localhost:{}/basic/string".format(port))
    policies = [UserAgentPolicy("myusergant"), AsyncRedirectPolicy()]
    async with AsyncPipeline(AsyncioRequestsTransport(), policies=policies) as pipeline:
        response = await pipeline.run(request)

    assert isinstance(response.http_response.status_code, int)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_conf_async_trio_requests(port, http_request):
    async def do():
        request = http_request("GET", "http://localhost:{}/basic/string".format(port))
        policies = [UserAgentPolicy("myusergant"), AsyncRedirectPolicy()]
        async with AsyncPipeline(TrioRequestsTransport(), policies=policies) as pipeline:
            return await pipeline.run(request)

    response = trio.run(do)
    assert isinstance(response.http_response.status_code, int)


@pytest.mark.trio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_conf_async_trio_auth_policy_concurrent(port, http_request):

    request = http_request("GET", "http://localhost:{}/basic/string".format(port))

    async def get_token(*_, **__):
        await trio.sleep(1)
        return AccessToken("token", 2524608000)

    async def run(p):
        response = await p.run(request, enforce_https=False)
        assert isinstance(response.http_response.status_code, int)

    fake_credential = Mock(get_token=get_token)
    policies = [AsyncBearerTokenCredentialPolicy(fake_credential, "scope")]
    async with AsyncPipeline(TrioRequestsTransport(), policies=policies) as pipeline, trio.open_nursery() as nursery:
        nursery.start_soon(run, pipeline)
        nursery.start_soon(run, pipeline)


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_retry_without_http_response(http_request):
    class NaughtyPolicy(AsyncHTTPPolicy):
        def send(*args):
            raise AzureError("boo")

    policies = [AsyncRetryPolicy(), NaughtyPolicy()]
    pipeline = AsyncPipeline(policies=policies, transport=None)
    with pytest.raises(AzureError):
        await pipeline.run(http_request("GET", url="https://foo.bar"))


@pytest.mark.asyncio
async def test_add_custom_policy():
    class BooPolicy(AsyncHTTPPolicy):
        def send(*args):
            raise AzureError("boo")

    class FooPolicy(AsyncHTTPPolicy):
        def send(*args):
            raise AzureError("boo")

    config = Configuration()
    retry_policy = AsyncRetryPolicy()
    config.retry_policy = retry_policy
    boo_policy = BooPolicy()
    foo_policy = FooPolicy()
    client = AsyncPipelineClient(base_url="test", config=config, per_call_policies=boo_policy)
    policies = client._pipeline._impl_policies
    assert boo_policy in policies
    pos_boo = policies.index(boo_policy)
    pos_retry = policies.index(retry_policy)
    assert pos_boo < pos_retry

    client = AsyncPipelineClient(base_url="test", config=config, per_call_policies=[boo_policy])
    policies = client._pipeline._impl_policies
    assert boo_policy in policies
    pos_boo = policies.index(boo_policy)
    pos_retry = policies.index(retry_policy)
    assert pos_boo < pos_retry

    client = AsyncPipelineClient(base_url="test", config=config, per_retry_policies=boo_policy)
    policies = client._pipeline._impl_policies
    assert boo_policy in policies
    pos_boo = policies.index(boo_policy)
    pos_retry = policies.index(retry_policy)
    assert pos_boo > pos_retry

    client = AsyncPipelineClient(base_url="test", config=config, per_retry_policies=[boo_policy])
    policies = client._pipeline._impl_policies
    assert boo_policy in policies
    pos_boo = policies.index(boo_policy)
    pos_retry = policies.index(retry_policy)
    assert pos_boo > pos_retry

    client = AsyncPipelineClient(
        base_url="test",
        config=config,
        per_call_policies=boo_policy,
        per_retry_policies=foo_policy,
    )
    policies = client._pipeline._impl_policies
    assert boo_policy in policies
    assert foo_policy in policies
    pos_boo = policies.index(boo_policy)
    pos_foo = policies.index(foo_policy)
    pos_retry = policies.index(retry_policy)
    assert pos_boo < pos_retry
    assert pos_foo > pos_retry

    client = AsyncPipelineClient(
        base_url="test",
        config=config,
        per_call_policies=[boo_policy],
        per_retry_policies=[foo_policy],
    )
    policies = client._pipeline._impl_policies
    assert boo_policy in policies
    assert foo_policy in policies
    pos_boo = policies.index(boo_policy)
    pos_foo = policies.index(foo_policy)
    pos_retry = policies.index(retry_policy)
    assert pos_boo < pos_retry
    assert pos_foo > pos_retry

    policies = [UserAgentPolicy(), AsyncRetryPolicy(), DistributedTracingPolicy()]
    client = AsyncPipelineClient(base_url="test", policies=policies, per_call_policies=boo_policy)
    actual_policies = client._pipeline._impl_policies
    assert boo_policy == actual_policies[0]
    client = AsyncPipelineClient(base_url="test", policies=policies, per_call_policies=[boo_policy])
    actual_policies = client._pipeline._impl_policies
    assert boo_policy == actual_policies[0]

    client = AsyncPipelineClient(base_url="test", policies=policies, per_retry_policies=foo_policy)
    actual_policies = client._pipeline._impl_policies
    assert foo_policy == actual_policies[2]
    client = AsyncPipelineClient(base_url="test", policies=policies, per_retry_policies=[foo_policy])
    actual_policies = client._pipeline._impl_policies
    assert foo_policy == actual_policies[2]

    client = AsyncPipelineClient(
        base_url="test",
        policies=policies,
        per_call_policies=boo_policy,
        per_retry_policies=[foo_policy],
    )
    actual_policies = client._pipeline._impl_policies
    assert boo_policy == actual_policies[0]
    assert foo_policy == actual_policies[3]
    client = AsyncPipelineClient(
        base_url="test",
        policies=policies,
        per_call_policies=[boo_policy],
        per_retry_policies=[foo_policy],
    )
    actual_policies = client._pipeline._impl_policies
    assert boo_policy == actual_policies[0]
    assert foo_policy == actual_policies[3]

    policies = [UserAgentPolicy(), DistributedTracingPolicy()]
    with pytest.raises(ValueError):
        client = AsyncPipelineClient(base_url="test", policies=policies, per_retry_policies=foo_policy)
    with pytest.raises(ValueError):
        client = AsyncPipelineClient(base_url="test", policies=policies, per_retry_policies=[foo_policy])


def test_no_cleanup_policy_when_redirect_policy_is_empty():
    config = Configuration()
    client = AsyncPipelineClient(base_url="test", config=config)
    policies = client._pipeline._impl_policies
    for policy in policies:
        if isinstance(policy, SensitiveHeaderCleanupPolicy):
            assert False


@pytest.mark.skipif(sys.version_info < (3, 8), reason="Python 3.7 does not support AsyncMock")
@pytest.mark.asyncio
async def test_default_ssl_context():
    class MockAiohttpSession:
        async def __aenter__(self):
            pass

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            pass

        async def close(self):
            pass

        async def open(self):
            pass

        async def request(self, method: str, url: str, **kwargs):
            assert "ssl" not in kwargs
            mock_response = AsyncMock(spec=aiohttp.ClientResponse)
            type(mock_response).status = PropertyMock(return_value=200)
            return mock_response

    transport = AioHttpTransport(session=MockAiohttpSession(), session_owner=False)
    pipeline = AsyncPipeline(transport=transport)

    req = HttpRequest("GET", "https://bing.com")
    await pipeline.run(req)



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import sys
import pytest
import pytest_asyncio
from azure.core.pipeline.transport import HttpRequest as PipelineTransportHttpRequest
from azure.core.rest import HttpRequest as RestHttpRequest
from azure.core.pipeline import Pipeline
from azure.core.pipeline.transport import AioHttpTransport, AsyncioRequestsTransport, TrioRequestsTransport
from rest_client_async import AsyncMockRestClient

TRANSPORTS = [AioHttpTransport, AsyncioRequestsTransport]


@pytest.fixture
def old_request(port):
    return PipelineTransportHttpRequest("GET", "http://localhost:{}/streams/basic".format(port))


@pytest_asyncio.fixture
async def get_old_response(old_request):
    async def _callback(transport, **kwargs):
        async with transport() as sender:
            return await sender.send(old_request, **kwargs)

    return _callback


@pytest.fixture
@pytest.mark.trio
async def get_old_response_trio(old_request):
    async def _callback(**kwargs):
        async with TrioRequestsTransport() as sender:
            return await sender.send(old_request, **kwargs)

    return _callback


@pytest.fixture
def new_request(port):
    return RestHttpRequest("GET", "http://localhost:{}/streams/basic".format(port))


@pytest_asyncio.fixture
async def get_new_response(new_request):
    async def _callback(transport, **kwargs):
        async with transport() as sender:
            return await sender.send(new_request, **kwargs)

    return _callback


@pytest.fixture
@pytest.mark.trio
async def get_new_response_trio(new_request):
    async def _callback(**kwargs):
        async with TrioRequestsTransport() as sender:
            return await sender.send(new_request, **kwargs)

    return _callback


def _test_response_attr_parity(old_response, new_response):
    for attr in dir(old_response):
        if not attr[0] == "_":
            # if not a private attr, we want parity
            assert hasattr(new_response, attr)


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_attr_parity(get_old_response, get_new_response, transport):
    old_response = await get_old_response(transport)
    new_response = await get_new_response(transport)
    _test_response_attr_parity(old_response, new_response)


@pytest.mark.trio
async def test_response_attr_parity_trio(get_old_response_trio, get_new_response_trio):
    old_response = await get_old_response_trio()
    new_response = await get_new_response_trio()
    _test_response_attr_parity(old_response, new_response)


def _test_response_set_attrs(old_response, new_response):
    for attr in dir(old_response):
        if attr[0] == "_":
            continue
        try:
            # if we can set it on the old request, we want to
            # be able to set it on the new
            setattr(old_response, attr, "foo")
        except:
            pass
        else:
            setattr(new_response, attr, "foo")
            assert getattr(old_response, attr) == getattr(new_response, attr) == "foo"


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_set_attrs(get_old_response, get_new_response, transport):
    old_response = await get_old_response(transport)
    new_response = await get_new_response(transport)
    _test_response_set_attrs(old_response, new_response)


@pytest.mark.trio
async def test_response_set_attrs_trio(get_old_response_trio, get_new_response_trio):
    old_response = await get_old_response_trio()
    new_response = await get_new_response_trio()
    _test_response_set_attrs(old_response, new_response)


def _test_response_block_size(old_response, new_response):
    assert old_response.block_size == new_response.block_size == 4096
    old_response.block_size = 500
    new_response.block_size = 500
    assert old_response.block_size == new_response.block_size == 500


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_block_size(get_old_response, get_new_response, transport):
    old_response = await get_old_response(transport)
    new_response = await get_new_response(transport)
    _test_response_block_size(old_response, new_response)


@pytest.mark.trio
async def test_response_block_size_trio(get_old_response_trio, get_new_response_trio):
    old_response = await get_old_response_trio()
    new_response = await get_new_response_trio()
    _test_response_block_size(old_response, new_response)


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_body(get_old_response, get_new_response, transport):
    old_response = await get_old_response(transport)
    new_response = await get_new_response(transport)
    assert old_response.body() == new_response.body() == b"Hello, world!"


@pytest.mark.trio
async def test_response_body_trio(get_old_response_trio, get_new_response_trio):
    old_response = await get_old_response_trio()
    new_response = await get_new_response_trio()
    assert old_response.body() == new_response.body() == b"Hello, world!"


def _test_response_internal_response(old_response, new_response, port):
    assert (
        str(old_response.internal_response.url)
        == str(new_response.internal_response.url)
        == "http://localhost:{}/streams/basic".format(port)
    )
    old_response.internal_response = "foo"
    new_response.internal_response = "foo"
    assert old_response.internal_response == new_response.internal_response == "foo"


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_internal_response(get_old_response, get_new_response, transport, port):
    old_response = await get_old_response(transport)
    new_response = await get_new_response(transport)
    _test_response_internal_response(old_response, new_response, port)


@pytest.mark.trio
async def test_response_internal_response_trio(get_old_response_trio, get_new_response_trio, port):
    old_response = await get_old_response_trio()
    new_response = await get_new_response_trio()
    _test_response_internal_response(old_response, new_response, port)


@pytest.mark.skip(reason="Resolve in #32162")
@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_stream_download(get_old_response, get_new_response, transport):
    old_response = await get_old_response(transport, stream=True)
    new_response = await get_new_response(transport, stream=True)
    pipeline = Pipeline(transport())
    old_string = b"".join([part async for part in old_response.stream_download(pipeline=pipeline)])
    new_string = b"".join([part async for part in new_response.stream_download(pipeline=pipeline)])

    # aiohttp can be flaky for both old and new responses, so since we're just checking backcompat here
    # using in instead of equals
    assert old_string in b"Hello, world!"
    assert new_string in b"Hello, world!"


@pytest.mark.trio
async def test_response_stream_download_trio(get_old_response_trio, get_new_response_trio):
    old_response = await get_old_response_trio(stream=True)
    new_response = await get_new_response_trio(stream=True)
    pipeline = Pipeline(TrioRequestsTransport())
    old_string = b"".join([part async for part in old_response.stream_download(pipeline=pipeline)])
    new_string = b"".join([part async for part in new_response.stream_download(pipeline=pipeline)])
    assert old_string == new_string == b"Hello, world!"


def _test_response_request(old_response, new_response, port):
    assert old_response.request.url == new_response.request.url == "http://localhost:{}/streams/basic".format(port)
    old_response.request = "foo"
    new_response.request = "foo"
    assert old_response.request == new_response.request == "foo"


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_request(get_old_response, get_new_response, port, transport):
    old_response = await get_old_response(transport)
    new_response = await get_new_response(transport)
    _test_response_request(old_response, new_response, port)


@pytest.mark.trio
async def test_response_request_trio(get_old_response_trio, get_new_response_trio, port):
    old_response = await get_old_response_trio()
    new_response = await get_new_response_trio()
    _test_response_request(old_response, new_response, port)


def _test_response_status_code(old_response, new_response):
    assert old_response.status_code == new_response.status_code == 200
    old_response.status_code = 202
    new_response.status_code = 202
    assert old_response.status_code == new_response.status_code == 202


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_status_code(get_old_response, get_new_response, transport):
    old_response = await get_old_response(transport)
    new_response = await get_new_response(transport)
    _test_response_status_code(old_response, new_response)


@pytest.mark.trio
async def test_response_status_code_trio(get_old_response_trio, get_new_response_trio):
    old_response = await get_old_response_trio()
    new_response = await get_new_response_trio()
    _test_response_status_code(old_response, new_response)


def _test_response_headers(old_response, new_response):
    assert (
        set(old_response.headers.keys())
        == set(new_response.headers.keys())
        == set(["Content-Type", "Connection", "Server", "Date", "Transfer-Encoding"])
    )
    old_response.headers = {"Hello": "world!"}
    new_response.headers = {"Hello": "world!"}
    assert old_response.headers == new_response.headers == {"Hello": "world!"}


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_headers(get_old_response, get_new_response, transport):
    old_response = await get_old_response(transport)
    new_response = await get_new_response(transport)
    _test_response_headers(old_response, new_response)


@pytest.mark.trio
async def test_response_headers_trio(get_old_response_trio, get_new_response_trio):
    old_response = await get_old_response_trio()
    new_response = await get_new_response_trio()
    _test_response_headers(old_response, new_response)


def _test_response_reason(old_response, new_response):
    assert old_response.reason == new_response.reason == "OK"
    old_response.reason = "Not OK"
    new_response.reason = "Not OK"
    assert old_response.reason == new_response.reason == "Not OK"


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_reason(get_old_response, get_new_response, transport):
    old_response = await get_old_response(transport)
    new_response = await get_new_response(transport)
    _test_response_reason(old_response, new_response)


@pytest.mark.trio
async def test_response_reason_trio(get_old_response_trio, get_new_response_trio):
    old_response = await get_old_response_trio()
    new_response = await get_new_response_trio()
    _test_response_reason(old_response, new_response)


def _test_response_content_type(old_response, new_response):
    assert old_response.content_type == new_response.content_type == "text/html; charset=utf-8"
    old_response.content_type = "application/json"
    new_response.content_type = "application/json"
    assert old_response.content_type == new_response.content_type == "application/json"


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_content_type(get_old_response, get_new_response, transport):
    old_response = await get_old_response(transport)
    new_response = await get_new_response(transport)
    _test_response_content_type(old_response, new_response)


@pytest.mark.trio
async def test_response_content_type_trio(get_old_response_trio, get_new_response_trio):
    old_response = await get_old_response_trio()
    new_response = await get_new_response_trio()
    _test_response_content_type(old_response, new_response)


def _create_multiapart_request(http_request_class):
    class ResponsePolicy(object):
        def on_request(self, *args):
            return

        def on_response(self, request, response):
            response.http_response.headers["x-ms-fun"] = "true"

    class AsyncResponsePolicy(object):
        def on_request(self, *args):
            return

        async def on_response(self, request, response):
            response.http_response.headers["x-ms-async-fun"] = "true"

    req0 = http_request_class("DELETE", "/container0/blob0")
    req1 = http_request_class("DELETE", "/container1/blob1")
    request = http_request_class("POST", "/multipart/request")
    request.set_multipart_mixed(req0, req1, policies=[ResponsePolicy(), AsyncResponsePolicy()])
    return request


async def _test_parts(response):
    # hack the content type
    parts = [p async for p in response.parts()]
    assert len(parts) == 2

    parts0 = parts[0]
    assert parts0.status_code == 202
    assert parts0.headers["x-ms-fun"] == "true"
    assert parts0.headers["x-ms-async-fun"] == "true"

    parts1 = parts[1]
    assert parts1.status_code == 404
    assert parts1.headers["x-ms-fun"] == "true"
    assert parts1.headers["x-ms-async-fun"] == "true"


@pytest.mark.asyncio
@pytest.mark.parametrize("transport", TRANSPORTS)
async def test_response_parts(port, transport):
    # there's no support for trio + multipart rn
    old_request = _create_multiapart_request(PipelineTransportHttpRequest)
    new_request = _create_multiapart_request(RestHttpRequest)
    old_response = await AsyncMockRestClient(port, transport=transport()).send_request(old_request, stream=True)
    new_response = await AsyncMockRestClient(port, transport=transport()).send_request(new_request, stream=True)
    if hasattr(old_response, "load_body"):
        # only aiohttp has this attr
        await old_response.load_body()
        await new_response.load_body()
    await _test_parts(old_response)
    await _test_parts(new_response)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import pytest
import types

############################## LISTS USED TO PARAMETERIZE TESTS ##############################
from azure.core.rest import HttpRequest as RestHttpRequest
from azure.core.pipeline.transport import HttpRequest as PipelineTransportHttpRequest
from azure.core.pipeline._tools import is_rest

HTTP_REQUESTS = [PipelineTransportHttpRequest, RestHttpRequest]
REQUESTS_TRANSPORT_RESPONSES = []

from azure.core.pipeline.transport import HttpResponse as PipelineTransportHttpResponse
from azure.core.rest._http_response_impl import HttpResponseImpl as RestHttpResponse

HTTP_RESPONSES = [PipelineTransportHttpResponse, RestHttpResponse]

ASYNC_HTTP_RESPONSES = []

try:
    from azure.core.pipeline.transport import AsyncHttpResponse as PipelineTransportAsyncHttpResponse
    from azure.core.rest._http_response_impl_async import AsyncHttpResponseImpl as RestAsyncHttpResponse

    ASYNC_HTTP_RESPONSES = [PipelineTransportAsyncHttpResponse, RestAsyncHttpResponse]
except (ImportError, SyntaxError):
    pass

try:
    from azure.core.pipeline.transport import RequestsTransportResponse as PipelineTransportRequestsTransportResponse
    from azure.core.rest._requests_basic import RestRequestsTransportResponse

    REQUESTS_TRANSPORT_RESPONSES = [PipelineTransportRequestsTransportResponse, RestRequestsTransportResponse]
except ImportError:
    pass

from azure.core.pipeline.transport._base import (
    HttpClientTransportResponse as PipelineTransportHttpClientTransportResponse,
)
from azure.core.rest._http_response_impl import RestHttpClientTransportResponse

HTTP_CLIENT_TRANSPORT_RESPONSES = [PipelineTransportHttpClientTransportResponse, RestHttpClientTransportResponse]

ASYNCIO_REQUESTS_TRANSPORT_RESPONSES = []

try:
    from azure.core.pipeline.transport import (
        AsyncioRequestsTransportResponse as PipelineTransportAsyncioRequestsTransportResponse,
    )
    from azure.core.rest._requests_asyncio import RestAsyncioRequestsTransportResponse

    ASYNCIO_REQUESTS_TRANSPORT_RESPONSES = [
        PipelineTransportAsyncioRequestsTransportResponse,
        RestAsyncioRequestsTransportResponse,
    ]
except (ImportError, SyntaxError):
    pass

AIOHTTP_TRANSPORT_RESPONSES = []

try:
    from azure.core.pipeline.transport import AioHttpTransportResponse as PipelineTransportAioHttpTransportResponse
    from azure.core.rest._aiohttp import RestAioHttpTransportResponse

    AIOHTTP_TRANSPORT_RESPONSES = [PipelineTransportAioHttpTransportResponse, RestAioHttpTransportResponse]
except (ImportError, SyntaxError):
    pass

############################## HELPER FUNCTIONS ##############################


def request_and_responses_product(*args):
    pipeline_transport = tuple([PipelineTransportHttpRequest]) + tuple(arg[0] for arg in args)
    rest = tuple([RestHttpRequest]) + tuple(arg[1] for arg in args)
    return [pipeline_transport, rest]


def create_http_request(http_request, *args, **kwargs):
    if hasattr(http_request, "content"):
        method = args[0]
        url = args[1]
        try:
            headers = args[2]
        except IndexError:
            headers = None
        try:
            files = args[3]
        except IndexError:
            files = None
        try:
            data = args[4]
        except IndexError:
            data = None
        return http_request(method=method, url=url, headers=headers, files=files, data=data, **kwargs)
    return http_request(*args, **kwargs)


def create_transport_response(http_response, *args, **kwargs):
    # this creates transport-specific responses,
    # like requests responses / aiohttp responses
    if is_rest(http_response):
        block_size = args[2] if len(args) > 2 else None
        return http_response(request=args[0], internal_response=args[1], block_size=block_size, **kwargs)
    return http_response(*args, **kwargs)


def create_http_response(http_response, *args, **kwargs):
    # since the actual http_response object is
    # an ABC for our new responses, it's a little more
    # complicated creating a pure http response.
    # here, we use the HttpResponsImpl, but we still have to pass
    # more things to the init, so making a separate operation
    if is_rest(http_response):
        block_size = args[2] if len(args) > 2 else None
        return http_response(
            request=args[0],
            internal_response=args[1],
            block_size=block_size,
            status_code=kwargs.pop("status_code", 200),
            reason=kwargs.pop("reason", "OK"),
            content_type=kwargs.pop("content_type", "application/json"),
            headers=kwargs.pop("headers", {}),
            stream_download_generator=kwargs.pop("stream_download_generator", None),
            **kwargs
        )
    return http_response(*args, **kwargs)


def readonly_checks(response, old_response_class):
    # though we want these properties to be completely readonly, it doesn't work
    # for the backcompat properties
    assert isinstance(response.request, RestHttpRequest)
    assert isinstance(response.status_code, int)
    assert response.headers
    assert response.content_type == "text/html; charset=utf-8"

    assert response.is_closed
    with pytest.raises(AttributeError):
        response.is_closed = False

    assert response.is_stream_consumed
    with pytest.raises(AttributeError):
        response.is_stream_consumed = False

    # you can set encoding
    assert response.encoding == "utf-8"
    response.encoding = "blah"
    assert response.encoding == "blah"

    assert isinstance(response.url, str)
    with pytest.raises(AttributeError):
        response.url = "http://fakeurl"

    assert response.content is not None
    with pytest.raises(AttributeError):
        response.content = b"bad"

    old_response = old_response_class(response.request, response.internal_response, response.block_size)
    for attr in dir(response):
        if attr[0] == "_":
            # don't care about private variables
            continue
        if type(getattr(response, attr)) == types.MethodType:
            # methods aren't "readonly"
            continue
        if attr == "encoding":
            # encoding is the only settable new attr
            continue
        if not attr in vars(old_response):
            with pytest.raises(AttributeError):
                setattr(response, attr, "new_value")



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
"""The tests for decorators.py and common.py"""

try:
    from unittest import mock
except ImportError:
    import mock

import time

import pytest
from azure.core.pipeline import Pipeline, PipelineResponse
from azure.core.pipeline.policies import HTTPPolicy
from azure.core.pipeline.transport import HttpTransport
from azure.core.settings import settings
from azure.core.tracing import common, SpanKind
from azure.core.tracing.decorator import distributed_trace
from tracing_common import FakeSpan
from utils import HTTP_REQUESTS


@pytest.fixture(scope="module")
def fake_span():
    settings.tracing_implementation.set_value(FakeSpan)


class MockClient:
    @distributed_trace
    def __init__(self, http_request, policies=None, assert_current_span=False):
        time.sleep(0.001)
        self.request = http_request("GET", "http://localhost")
        if policies is None:
            policies = []
        policies.append(mock.Mock(spec=HTTPPolicy, send=self.verify_request))
        self.policies = policies
        self.transport = mock.Mock(spec=HttpTransport)
        self.pipeline = Pipeline(self.transport, policies=policies)

        self.expected_response = mock.Mock(spec=PipelineResponse)
        self.assert_current_span = assert_current_span

    def verify_request(self, request):
        if self.assert_current_span:
            assert execution_context.get_current_span() is not None
        return self.expected_response

    @distributed_trace
    def make_request(self, numb_times, **kwargs):
        time.sleep(0.001)
        if numb_times < 1:
            return None
        response = self.pipeline.run(self.request, **kwargs)
        self.get_foo(merge_span=True)
        kwargs["merge_span"] = True
        self.make_request(numb_times - 1, **kwargs)
        return response

    @distributed_trace
    def merge_span_method(self):
        return self.get_foo(merge_span=True)

    @distributed_trace
    def no_merge_span_method(self):
        return self.get_foo()

    @distributed_trace
    def get_foo(self):
        time.sleep(0.001)
        return 5

    @distributed_trace(name_of_span="different name")
    def check_name_is_different(self):
        time.sleep(0.001)

    @distributed_trace(tracing_attributes={"foo": "bar"})
    def tracing_attr(self):
        time.sleep(0.001)

    @distributed_trace(kind=SpanKind.PRODUCER)
    def kind_override(self):
        time.sleep(0.001)

    @distributed_trace
    def raising_exception(self):
        raise ValueError("Something went horribly wrong here")


def random_function():
    pass


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_get_function_and_class_name(http_request):
    client = MockClient(http_request)
    assert common.get_function_and_class_name(client.get_foo, client) == "MockClient.get_foo"
    assert common.get_function_and_class_name(random_function) == "random_function"


@pytest.mark.usefixtures("fake_span")
class TestDecorator(object):
    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_decorator_tracing_attr(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            client.tracing_attr()

        assert len(parent.children) == 2
        assert parent.children[0].name == "MockClient.__init__"
        assert parent.children[1].name == "MockClient.tracing_attr"
        assert parent.children[1].kind == SpanKind.INTERNAL
        assert parent.children[1].attributes == {"foo": "bar"}

    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_decorator_has_different_name(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            client.check_name_is_different()

        assert len(parent.children) == 2
        assert parent.children[0].name == "MockClient.__init__"
        assert parent.children[1].name == "different name"
        assert parent.children[1].kind == SpanKind.INTERNAL

    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_kind_override(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            client.kind_override()

        assert len(parent.children) == 2
        assert parent.children[0].name == "MockClient.__init__"
        assert parent.children[1].name == "MockClient.kind_override"
        assert parent.children[1].kind == SpanKind.PRODUCER

    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_used(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request, policies=[])
            client.get_foo(parent_span=parent)
            client.get_foo()

        assert len(parent.children) == 3
        assert parent.children[0].name == "MockClient.__init__"
        assert not parent.children[0].children
        assert parent.children[1].name == "MockClient.get_foo"
        assert not parent.children[1].children
        assert parent.children[2].name == "MockClient.get_foo"
        assert not parent.children[2].children

    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_span_merge_span(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            client.merge_span_method()
            client.no_merge_span_method()

        assert len(parent.children) == 3
        assert parent.children[0].name == "MockClient.__init__"
        assert not parent.children[0].children
        assert parent.children[1].name == "MockClient.merge_span_method"
        assert not parent.children[1].children
        assert parent.children[2].name == "MockClient.no_merge_span_method"
        assert parent.children[2].children[0].name == "MockClient.get_foo"

    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_span_complicated(self, http_request):
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            client.make_request(2)
            with parent.span("child") as child:
                time.sleep(0.001)
                client.make_request(2, parent_span=parent)
                assert FakeSpan.get_current_span() == child
                client.make_request(2)

        assert len(parent.children) == 4
        assert parent.children[0].name == "MockClient.__init__"
        assert not parent.children[0].children
        assert parent.children[1].name == "MockClient.make_request"
        assert not parent.children[1].children
        assert parent.children[2].name == "child"
        assert parent.children[2].children[0].name == "MockClient.make_request"
        assert parent.children[3].name == "MockClient.make_request"
        assert not parent.children[3].children

    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_span_with_exception(self, http_request):
        """Assert that if an exception is raised, the next sibling method is actually a sibling span."""
        with FakeSpan(name="parent") as parent:
            client = MockClient(http_request)
            try:
                client.raising_exception()
            except:
                pass
            client.get_foo()

        assert len(parent.children) == 3
        assert parent.children[0].name == "MockClient.__init__"
        assert parent.children[1].name == "MockClient.raising_exception"
        # Exception should propagate status for Opencensus
        assert parent.children[1].status == "Something went horribly wrong here"
        assert parent.children[2].name == "MockClient.get_foo"



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
import pytest
import json
import requests
from unittest.mock import Mock

# module under test
from azure.core.exceptions import (
    HttpResponseError,
    ODataV4Error,
    ODataV4Format,
    SerializationError,
    DeserializationError,
)
from azure.core.pipeline.transport import RequestsTransportResponse
from azure.core.pipeline.transport._base import _HttpResponseBase as PipelineTransportHttpResponseBase
from azure.core.rest._http_response_impl import _HttpResponseBaseImpl as RestHttpResponseBase
from utils import HTTP_REQUESTS


class PipelineTransportMockResponse(PipelineTransportHttpResponseBase):
    def __init__(self, json_body):
        super(PipelineTransportMockResponse, self).__init__(
            request=None,
            internal_response=None,
        )
        self.status_code = 400
        self.reason = "Bad Request"
        self.content_type = "application/json"
        self._body = json_body

    def body(self):
        return self._body


class RestMockResponse(RestHttpResponseBase):
    def __init__(self, json_body):
        super(RestMockResponse, self).__init__(
            request=None,
            internal_response=None,
            status_code=400,
            reason="Bad Request",
            content_type="application/json",
            headers={},
            stream_download_generator=None,
        )
        self._body = json_body

    def body(self):
        return self._body

    @property
    def content(self):
        return self._body


MOCK_RESPONSES = [PipelineTransportMockResponse, RestMockResponse]


class FakeErrorOne(object):
    def __init__(self):
        self.error = Mock(message="A fake error", code="FakeErrorOne")


class FakeErrorTwo(object):
    def __init__(self):
        self.code = "FakeErrorTwo"
        self.message = "A different fake error"


class FakeHttpResponse(HttpResponseError):
    def __init__(self, response, error, *args, **kwargs):
        self.error = error
        super(FakeHttpResponse, self).__init__(self, response=response, *args, **kwargs)


class TestExceptions(object):
    def test_empty_httpresponse_error(self):
        error = HttpResponseError()
        assert str(error) == "Operation returned an invalid status 'None'"
        assert error.message == "Operation returned an invalid status 'None'"
        assert error.response is None
        assert error.reason is None
        assert error.error is None
        assert error.status_code is None

    def test_message_httpresponse_error(self):
        error = HttpResponseError(message="Specific error message")
        assert str(error) == "Specific error message"
        assert error.message == "Specific error message"
        assert error.response is None
        assert error.reason is None
        assert error.error is None
        assert error.status_code is None

    def test_error_continuation_token(self):
        error = HttpResponseError(message="Specific error message", continuation_token="foo")
        assert str(error) == "Specific error message"
        assert error.message == "Specific error message"
        assert error.response is None
        assert error.reason is None
        assert error.error is None
        assert error.status_code is None
        assert error.continuation_token == "foo"

    @pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
    def test_deserialized_httpresponse_error_code(self, mock_response):
        """This is backward compat support of autorest azure-core (KV 4.0.0, Storage 12.0.0).

        Do NOT adapt this test unless you know what you're doing.
        """
        message = {
            "error": {
                "code": "FakeErrorOne",
                "message": "A fake error",
            }
        }
        response = mock_response(json.dumps(message).encode("utf-8"))
        error = FakeHttpResponse(response, FakeErrorOne())
        assert "(FakeErrorOne) A fake error" in error.message
        assert "(FakeErrorOne) A fake error" in str(error.error)
        assert error.error.code == "FakeErrorOne"
        assert error.error.message == "A fake error"
        assert error.response is response
        assert error.reason == "Bad Request"
        assert error.status_code == 400
        assert isinstance(error.model, FakeErrorOne)
        assert isinstance(error.error, ODataV4Format)

        # Could test if we see a deprecation warning
        assert error.error.error.code == "FakeErrorOne"
        assert error.error.error.message == "A fake error"

        assert str(error) == "(FakeErrorOne) A fake error\nCode: FakeErrorOne\nMessage: A fake error"

    @pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
    def test_deserialized_httpresponse_error_message(self, mock_response):
        """This is backward compat support for weird responses, adn even if it's likely
        just the autorest testserver, should be fine parsing.

        Do NOT adapt this test unless you know what you're doing.
        """
        message = {
            "code": "FakeErrorTwo",
            "message": "A different fake error",
        }
        response = mock_response(json.dumps(message).encode("utf-8"))
        error = FakeHttpResponse(response, FakeErrorTwo())
        assert "(FakeErrorTwo) A different fake error" in error.message
        assert "(FakeErrorTwo) A different fake error" in str(error.error)
        assert error.error.code == "FakeErrorTwo"
        assert error.error.message == "A different fake error"
        assert error.response is response
        assert error.reason == "Bad Request"
        assert error.status_code == 400
        assert isinstance(error.model, FakeErrorTwo)
        assert isinstance(error.error, ODataV4Format)

    @pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
    def test_httpresponse_error_with_response(self, port, mock_response):
        response = requests.get("http://localhost:{}/basic/string".format(port))
        http_response = RequestsTransportResponse(None, response)

        error = HttpResponseError(response=http_response)
        assert error.message == "Operation returned an invalid status 'OK'"
        assert error.response is not None
        assert error.reason == "OK"
        assert isinstance(error.status_code, int)
        assert error.error is None

    @pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
    def test_odata_v4_exception(self, mock_response):
        message = {
            "error": {
                "code": "501",
                "message": "Unsupported functionality",
                "target": "query",
                "details": [
                    {
                        "code": "301",
                        "target": "$search",
                        "message": "$search query option not supported",
                    }
                ],
                "innererror": {"trace": [], "context": {}},
            }
        }
        exp = ODataV4Error(mock_response(json.dumps(message).encode("utf-8")))

        assert exp.code == "501"
        assert exp.message == "Unsupported functionality"
        assert exp.target == "query"
        assert exp.details[0].code == "301"
        assert exp.details[0].target == "$search"
        assert "trace" in exp.innererror
        assert "context" in exp.innererror

        message = {}
        exp = ODataV4Error(mock_response(json.dumps(message).encode("utf-8")))
        assert exp.message == "Operation returned an invalid status 'Bad Request'"

        exp = ODataV4Error(mock_response(b""))
        assert exp.message == "Operation returned an invalid status 'Bad Request'"
        assert str(exp) == "Operation returned an invalid status 'Bad Request'"

    @pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
    def test_odata_v4_minimal(self, mock_response):
        """Minimal valid OData v4 is code/message and nothing else."""
        message = {
            "error": {
                "code": "501",
                "message": "Unsupported functionality",
            }
        }
        exp = ODataV4Error(mock_response(json.dumps(message).encode("utf-8")))
        assert exp.code == "501"
        assert exp.message == "Unsupported functionality"
        assert exp.target is None
        assert exp.details == []
        assert exp.innererror == {}

    @pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
    def test_broken_odata_details(self, mock_response):
        """Do not block creating a nice exception if "details" only is broken"""
        message = {
            "error": {
                "code": "Conflict",
                "message": "The maximum number of Free ServerFarms allowed in a Subscription is 10.",
                "target": None,
                "details": [
                    {"message": "The maximum number of Free ServerFarms allowed in a Subscription is 10."},
                    {"code": "Conflict"},
                    {
                        "errorentity": {
                            "code": "Conflict",
                            "message": "The maximum number of Free ServerFarms allowed in a Subscription is 10.",
                            "extendedCode": "59301",
                            "messageTemplate": "The maximum number of {0} ServerFarms allowed in a Subscription is {1}.",
                            "parameters": ["Free", "10"],
                            "innerErrors": None,
                        }
                    },
                ],
                "innererror": None,
            }
        }
        exp = HttpResponseError(response=mock_response(json.dumps(message).encode("utf-8")))
        assert exp.error.code == "Conflict"

    @pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
    def test_null_odata_details(self, mock_response):
        message = {
            "error": {
                "code": "501",
                "message": "message",
                "target": None,
                "details": None,
                "innererror": None,
            }
        }
        exp = HttpResponseError(response=mock_response(json.dumps(message).encode("utf-8")))
        assert exp.error.code == "501"

    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_non_odatav4_error_body(self, client, http_request):
        request = http_request("GET", "/errors/non-odatav4-body")
        response = client.send_request(request)
        with pytest.raises(HttpResponseError) as ex:
            response.raise_for_status()
        assert (
            str(ex.value)
            == 'Operation returned an invalid status \'BAD REQUEST\'\nContent: {"code": 400, "error": {"global": ["MY-ERROR-MESSAGE-THAT-IS-COMING-FROM-THE-API"]}}'
        )

    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_malformed_json(self, client, http_request):
        request = http_request("GET", "/errors/malformed-json")
        response = client.send_request(request)
        with pytest.raises(HttpResponseError) as ex:
            response.raise_for_status()
        assert (
            str(ex.value)
            == 'Operation returned an invalid status \'BAD REQUEST\'\nContent: {"code": 400, "error": {"global": ["MY-ERROR-MESSAGE-THAT-IS-COMING-FROM-THE-API"]'
        )

    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_text(self, client, http_request):
        request = http_request("GET", "/errors/text")
        response = client.send_request(request)
        with pytest.raises(HttpResponseError) as ex:
            response.raise_for_status()
        assert str(ex.value) == "Operation returned an invalid status 'BAD REQUEST'\nContent: I am throwing an error"

    @pytest.mark.parametrize("http_request", HTTP_REQUESTS)
    def test_datav4_error(self, client, http_request):
        request = http_request("GET", "/errors/odatav4")
        response = client.send_request(request)
        with pytest.raises(HttpResponseError) as ex:
            response.raise_for_status()
        assert 'Content: {"' not in str(ex.value)


def test_serialization_error():
    message = "Oopsy bad input passed for serialization"
    error = SerializationError(message)
    with pytest.raises(SerializationError) as ex:
        raise error
    assert str(ex.value) == message

    with pytest.raises(ValueError) as ex:
        raise error
    assert str(ex.value) == message


def test_deserialization_error():
    message = "Oopsy bad input passed for serialization"
    error = DeserializationError(message)
    with pytest.raises(DeserializationError) as ex:
        raise error
    assert str(ex.value) == message

    with pytest.raises(ValueError) as ex:
        raise error
    assert str(ex.value) == message



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import asyncio
import time
from unittest.mock import Mock
from requests import Response

from azure.core.credentials import AccessToken
from azure.core.credentials_async import AsyncTokenCredential
from azure.core.exceptions import ServiceRequestError
from azure.core.pipeline import AsyncPipeline
from azure.core.pipeline.policies import (
    AsyncBearerTokenCredentialPolicy,
    SansIOHTTPPolicy,
    AsyncRedirectPolicy,
    SensitiveHeaderCleanupPolicy,
)
from azure.core.pipeline.transport import AsyncHttpTransport, HttpRequest
import pytest

pytestmark = pytest.mark.asyncio
from utils import HTTP_REQUESTS


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_bearer_policy_adds_header(http_request):
    """The bearer token policy should add a header containing a token from its credential"""
    # 2524608000 == 01/01/2050 @ 12:00am (UTC)
    expected_token = AccessToken("expected_token", 2524608000)

    async def verify_authorization_header(request):
        assert request.http_request.headers["Authorization"] == "Bearer {}".format(expected_token.token)
        return Mock()

    get_token_calls = 0

    async def get_token(*_, **__):
        nonlocal get_token_calls
        get_token_calls += 1
        return expected_token

    fake_credential = Mock(get_token=get_token)
    policies = [AsyncBearerTokenCredentialPolicy(fake_credential, "scope"), Mock(send=verify_authorization_header)]
    pipeline = AsyncPipeline(transport=Mock(), policies=policies)

    await pipeline.run(http_request("GET", "https://spam.eggs"), context=None)
    assert get_token_calls == 1

    await pipeline.run(http_request("GET", "https://spam.eggs"), context=None)
    # Didn't need a new token
    assert get_token_calls == 1


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_bearer_policy_send(http_request):
    """The bearer token policy should invoke the next policy's send method and return the result"""
    expected_request = http_request("GET", "https://spam.eggs")
    expected_response = Mock()

    async def verify_request(request):
        assert request.http_request is expected_request
        return expected_response

    get_token = get_completed_future(AccessToken("***", 42))
    fake_credential = Mock(get_token=lambda *_, **__: get_token)
    policies = [AsyncBearerTokenCredentialPolicy(fake_credential, "scope"), Mock(send=verify_request)]
    response = await AsyncPipeline(transport=Mock(), policies=policies).run(expected_request)

    assert response is expected_response


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_bearer_policy_sync_send(http_request):
    """The bearer token policy should invoke the next policy's send method and return the result"""
    expected_request = http_request("GET", "https://spam.eggs")
    expected_response = Mock()

    async def verify_request(request):
        assert request.http_request is expected_request
        return expected_response

    get_token = get_completed_future(AccessToken("***", 42))
    fake_credential = Mock(get_token=lambda *_, **__: get_token)
    policies = [AsyncBearerTokenCredentialPolicy(fake_credential, "scope"), Mock(send=verify_request)]
    response = await AsyncPipeline(transport=Mock(), policies=policies).run(expected_request)

    assert response is expected_response


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_bearer_policy_token_caching(http_request):
    good_for_one_hour = AccessToken("token", time.time() + 3600)
    expected_token = good_for_one_hour
    get_token_calls = 0

    async def get_token(*_, **__):
        nonlocal get_token_calls
        get_token_calls += 1
        return expected_token

    credential = Mock(get_token=get_token)
    policies = [
        AsyncBearerTokenCredentialPolicy(credential, "scope"),
        Mock(send=Mock(return_value=get_completed_future(Mock()))),
    ]
    pipeline = AsyncPipeline(transport=Mock, policies=policies)

    await pipeline.run(http_request("GET", "https://spam.eggs"))
    assert get_token_calls == 1  # policy has no token at first request -> it should call get_token

    await pipeline.run(http_request("GET", "https://spam.eggs"))
    assert get_token_calls == 1  # token is good for an hour -> policy should return it from cache

    expired_token = AccessToken("token", time.time())
    get_token_calls = 0
    expected_token = expired_token
    policies = [
        AsyncBearerTokenCredentialPolicy(credential, "scope"),
        Mock(send=lambda _: get_completed_future(Mock())),
    ]
    pipeline = AsyncPipeline(transport=Mock(), policies=policies)

    await pipeline.run(http_request("GET", "https://spam.eggs"))
    assert get_token_calls == 1

    await pipeline.run(http_request("GET", "https://spam.eggs"))
    assert get_token_calls == 2  # token expired -> policy should call get_token


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_bearer_policy_optionally_enforces_https(http_request):
    """HTTPS enforcement should be controlled by a keyword argument, and enabled by default"""

    async def assert_option_popped(request, **kwargs):
        assert "enforce_https" not in kwargs, "AsyncBearerTokenCredentialPolicy didn't pop the 'enforce_https' option"
        return Mock()

    credential = Mock(get_token=lambda *_, **__: get_completed_future(AccessToken("***", 42)))
    pipeline = AsyncPipeline(
        transport=Mock(send=assert_option_popped), policies=[AsyncBearerTokenCredentialPolicy(credential, "scope")]
    )

    # by default and when enforce_https=True, the policy should raise when given an insecure request
    with pytest.raises(ServiceRequestError):
        await pipeline.run(http_request("GET", "http://not.secure"))
    with pytest.raises(ServiceRequestError):
        await pipeline.run(http_request("GET", "http://not.secure"), enforce_https=True)

    # when enforce_https=False, an insecure request should pass
    await pipeline.run(http_request("GET", "http://not.secure"), enforce_https=False)

    # https requests should always pass
    await pipeline.run(http_request("GET", "https://secure"), enforce_https=False)
    await pipeline.run(http_request("GET", "https://secure"), enforce_https=True)
    await pipeline.run(http_request("GET", "https://secure"))


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_bearer_policy_preserves_enforce_https_opt_out(http_request):
    """The policy should use request context to preserve an opt out from https enforcement"""

    class ContextValidator(SansIOHTTPPolicy):
        def on_request(self, request):
            assert "enforce_https" in request.context, "'enforce_https' is not in the request's context"
            return Mock()

    get_token = get_completed_future(AccessToken("***", 42))
    credential = Mock(get_token=lambda *_, **__: get_token)
    policies = [AsyncBearerTokenCredentialPolicy(credential, "scope"), ContextValidator()]
    pipeline = AsyncPipeline(transport=Mock(send=lambda *_, **__: get_completed_future(Mock())), policies=policies)

    await pipeline.run(http_request("GET", "http://not.secure"), enforce_https=False)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_bearer_policy_context_unmodified_by_default(http_request):
    """When no options for the policy accompany a request, the policy shouldn't add anything to the request context"""

    class ContextValidator(SansIOHTTPPolicy):
        def on_request(self, request):
            assert not any(request.context), "the policy shouldn't add to the request's context"
            return Mock()

    get_token = get_completed_future(AccessToken("***", 42))
    credential = Mock(get_token=lambda *_, **__: get_token)
    policies = [AsyncBearerTokenCredentialPolicy(credential, "scope"), ContextValidator()]
    pipeline = AsyncPipeline(transport=Mock(send=lambda *_, **__: get_completed_future(Mock())), policies=policies)

    await pipeline.run(http_request("GET", "https://secure"))


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_bearer_policy_calls_sansio_methods(http_request):
    """AsyncBearerTokenCredentialPolicy should call SansIOHttpPolicy methods as does _SansIOAsyncHTTPPolicyRunner"""

    class TestPolicy(AsyncBearerTokenCredentialPolicy):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.on_exception = Mock(return_value=False)
            self.on_request = Mock()
            self.on_response = Mock()

        async def send(self, request):
            self.request = request
            self.response = await super().send(request)
            return self.response

    credential = Mock(get_token=Mock(return_value=get_completed_future(AccessToken("***", int(time.time()) + 3600))))
    policy = TestPolicy(credential, "scope")
    transport = Mock(send=Mock(return_value=get_completed_future(Mock(status_code=200))))

    pipeline = AsyncPipeline(transport=transport, policies=[policy])
    await pipeline.run(http_request("GET", "https://localhost"))

    policy.on_request.assert_called_once_with(policy.request)
    policy.on_response.assert_called_once_with(policy.request, policy.response)

    # the policy should call on_exception when next.send() raises
    class TestException(Exception):
        pass

    # during the first send...
    transport = Mock(send=Mock(side_effect=TestException))
    policy = TestPolicy(credential, "scope")
    pipeline = AsyncPipeline(transport=transport, policies=[policy])
    with pytest.raises(TestException):
        await pipeline.run(http_request("GET", "https://localhost"))
    policy.on_exception.assert_called_once_with(policy.request)

    # ...or the second
    async def fake_send(*args, **kwargs):
        if fake_send.calls == 0:
            fake_send.calls = 1
            return Mock(status_code=401, headers={"WWW-Authenticate": 'Basic realm="localhost"'})
        raise TestException()

    fake_send.calls = 0

    policy = TestPolicy(credential, "scope")
    policy.on_challenge = Mock(return_value=get_completed_future(True))
    transport = Mock(send=Mock(wraps=fake_send))
    pipeline = AsyncPipeline(transport=transport, policies=[policy])
    with pytest.raises(TestException):
        await pipeline.run(http_request("GET", "https://localhost"))
    assert transport.send.call_count == 2
    policy.on_challenge.assert_called_once()
    policy.on_exception.assert_called_once_with(policy.request)


def get_completed_future(result=None):
    fut = asyncio.Future()
    fut.set_result(result)
    return fut


@pytest.mark.asyncio
async def test_bearer_policy_redirect_same_domain():
    class MockTransport(AsyncHttpTransport):
        def __init__(self):
            self._first = True

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            pass

        async def close(self):
            pass

        async def open(self):
            pass

        async def send(self, request, **kwargs):  # type: (PipelineRequest, Any) -> PipelineResponse
            if self._first:
                self._first = False
                assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
                response = Response()
                response.status_code = 301
                response.headers["location"] = "https://localhost"
                return response
            assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
            response = Response()
            response.status_code = 200
            return response

    auth_headder = "token"
    expected_scope = "scope"

    async def get_token(*_, **__):
        token = AccessToken(auth_headder, 0)
        return token

    credential = Mock(get_token=get_token)
    auth_policy = AsyncBearerTokenCredentialPolicy(credential, expected_scope)
    redirect_policy = AsyncRedirectPolicy()
    header_clean_up_policy = SensitiveHeaderCleanupPolicy()
    pipeline = AsyncPipeline(transport=MockTransport(), policies=[redirect_policy, auth_policy, header_clean_up_policy])

    await pipeline.run(HttpRequest("GET", "https://localhost"))


@pytest.mark.asyncio
async def test_bearer_policy_redirect_different_domain():
    class MockTransport(AsyncHttpTransport):
        def __init__(self):
            self._first = True

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            pass

        async def close(self):
            pass

        async def open(self):
            pass

        async def send(self, request, **kwargs):  # type: (PipelineRequest, Any) -> PipelineResponse
            if self._first:
                self._first = False
                assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
                response = Response()
                response.status_code = 301
                response.headers["location"] = "https://localhost1"
                return response
            assert not request.headers.get("Authorization")
            response = Response()
            response.status_code = 200
            return response

    auth_headder = "token"
    expected_scope = "scope"

    async def get_token(*_, **__):
        token = AccessToken(auth_headder, 0)
        return token

    credential = Mock(get_token=get_token)
    auth_policy = AsyncBearerTokenCredentialPolicy(credential, expected_scope)
    redirect_policy = AsyncRedirectPolicy()
    header_clean_up_policy = SensitiveHeaderCleanupPolicy()
    pipeline = AsyncPipeline(transport=MockTransport(), policies=[redirect_policy, auth_policy, header_clean_up_policy])

    await pipeline.run(HttpRequest("GET", "https://localhost"))


@pytest.mark.asyncio
async def test_bearer_policy_redirect_opt_out_clean_up():
    class MockTransport(AsyncHttpTransport):
        def __init__(self):
            self._first = True

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            pass

        async def close(self):
            pass

        async def open(self):
            pass

        async def send(self, request, **kwargs):  # type: (PipelineRequest, Any) -> PipelineResponse
            if self._first:
                self._first = False
                assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
                response = Response()
                response.status_code = 301
                response.headers["location"] = "https://localhost1"
                return response
            assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
            response = Response()
            response.status_code = 200
            return response

    auth_headder = "token"
    expected_scope = "scope"

    async def get_token(*_, **__):
        token = AccessToken(auth_headder, 0)
        return token

    credential = Mock(get_token=get_token)
    auth_policy = AsyncBearerTokenCredentialPolicy(credential, expected_scope)
    redirect_policy = AsyncRedirectPolicy()
    header_clean_up_policy = SensitiveHeaderCleanupPolicy(disable_redirect_cleanup=True)
    pipeline = AsyncPipeline(transport=MockTransport(), policies=[redirect_policy, auth_policy, header_clean_up_policy])

    await pipeline.run(HttpRequest("GET", "https://localhost"))


@pytest.mark.asyncio
async def test_bearer_policy_redirect_customize_sensitive_headers():
    class MockTransport(AsyncHttpTransport):
        def __init__(self):
            self._first = True

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            pass

        async def close(self):
            pass

        async def open(self):
            pass

        async def send(self, request, **kwargs):  # type: (PipelineRequest, Any) -> PipelineResponse
            if self._first:
                self._first = False
                assert request.headers["Authorization"] == "Bearer {}".format(auth_headder)
                response = Response()
                response.status_code = 301
                response.headers["location"] = "https://localhost1"
                return response
            assert request.headers.get("Authorization")
            response = Response()
            response.status_code = 200
            return response

    auth_headder = "token"
    expected_scope = "scope"

    async def get_token(*_, **__):
        token = AccessToken(auth_headder, 0)
        return token

    credential = Mock(get_token=get_token)
    auth_policy = AsyncBearerTokenCredentialPolicy(credential, expected_scope)
    redirect_policy = AsyncRedirectPolicy()
    header_clean_up_policy = SensitiveHeaderCleanupPolicy(blocked_redirect_headers=["x-ms-authorization-auxiliary"])
    pipeline = AsyncPipeline(transport=MockTransport(), policies=[redirect_policy, auth_policy, header_clean_up_policy])

    await pipeline.run(HttpRequest("GET", "https://localhost"))


@pytest.mark.asyncio
async def test_async_token_credential_inheritance():
    class TestTokenCredential(AsyncTokenCredential):
        async def get_token(self, *scopes, **kwargs):
            return "TOKEN"

    cred = TestTokenCredential()
    await cred.get_token("scope")



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import requests
from azure.core.pipeline.transport import (
    AsyncHttpTransport,
    AsyncioRequestsTransportResponse,
    AioHttpTransport,
)
from azure.core.pipeline import AsyncPipeline, PipelineResponse
from azure.core.pipeline.transport._aiohttp import AioHttpStreamDownloadGenerator
from unittest import mock
import pytest
from utils import request_and_responses_product, ASYNC_HTTP_RESPONSES, create_http_response


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(ASYNC_HTTP_RESPONSES))
async def test_connection_error_response(http_request, http_response):
    class MockSession(object):
        def __init__(self):
            self.auto_decompress = True

        @property
        def auto_decompress(self):
            return self.auto_decompress

    class MockTransport(AsyncHttpTransport):
        def __init__(self):
            self._count = 0
            self.session = MockSession

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            pass

        async def close(self):
            pass

        async def open(self):
            pass

        async def send(self, request, **kwargs):
            request = http_request("GET", "http://localhost/")
            response = create_http_response(http_response, request, None)
            response.status_code = 200
            return response

    class MockContent:
        def __init__(self):
            self._first = True

        async def read(self, block_size):
            if self._first:
                self._first = False
                raise ConnectionError
            return None

    class MockInternalResponse:
        def __init__(self):
            self.headers = {}
            self.content = MockContent()

        async def close(self):
            pass

    class AsyncMock(mock.MagicMock):
        async def __call__(self, *args, **kwargs):
            return super(AsyncMock, self).__call__(*args, **kwargs)

    http_request = http_request("GET", "http://localhost/")
    pipeline = AsyncPipeline(MockTransport())
    http_response = create_http_response(http_response, http_request, None)
    http_response.internal_response = MockInternalResponse()
    stream = AioHttpStreamDownloadGenerator(pipeline, http_response, decompress=False)
    with mock.patch("asyncio.sleep", new_callable=AsyncMock):
        with pytest.raises(ConnectionError):
            await stream.__anext__()


@pytest.mark.asyncio
@pytest.mark.parametrize("http_response", ASYNC_HTTP_RESPONSES)
async def test_response_streaming_error_behavior(http_response):
    # Test to reproduce https://github.com/Azure/azure-sdk-for-python/issues/16723
    block_size = 103
    total_response_size = 500
    req_response = requests.Response()
    req_request = requests.Request()

    class FakeStreamWithConnectionError:
        # fake object for urllib3.response.HTTPResponse
        def __init__(self):
            self.total_response_size = 500

        def stream(self, chunk_size, decode_content=False):
            assert chunk_size == block_size
            left = total_response_size
            while left > 0:
                if left <= block_size:
                    raise requests.exceptions.ConnectionError()
                data = b"X" * min(chunk_size, left)
                left -= len(data)
                yield data

        def read(self, chunk_size, decode_content=False):
            assert chunk_size == block_size
            if self.total_response_size > 0:
                if self.total_response_size <= block_size:
                    raise requests.exceptions.ConnectionError()
                data = b"X" * min(chunk_size, self.total_response_size)
                self.total_response_size -= len(data)
                return data

        def close(self):
            pass

    req_response.raw = FakeStreamWithConnectionError()

    response = AsyncioRequestsTransportResponse(
        req_request,
        req_response,
        block_size,
    )

    async def mock_run(self, *args, **kwargs):
        return PipelineResponse(
            None,
            requests.Response(),
            None,
        )

    transport = AioHttpTransport()
    pipeline = AsyncPipeline(transport)
    pipeline.run = mock_run
    downloader = response.stream_download(pipeline)
    with pytest.raises(requests.exceptions.ConnectionError):
        while True:
            await downloader.__anext__()



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------


def test_example_raw_response_hook():
    def callback(response):
        response.http_response.status_code = 200
        response.http_response.headers["custom_header"] = "CustomHeader"

    from azure.core.pipeline import Pipeline
    from azure.core.rest import HttpRequest
    from azure.core.pipeline.policies import RedirectPolicy, UserAgentPolicy
    from azure.core.pipeline.transport import RequestsTransport
    from azure.core.pipeline.policies import CustomHookPolicy

    request = HttpRequest("GET", "https://bing.com")
    policies = [CustomHookPolicy(raw_response_hook=callback)]

    with Pipeline(transport=RequestsTransport(), policies=policies) as pipeline:
        response = pipeline.run(request)
        assert response.http_response.status_code == 200
        assert response.http_response.headers["custom_header"] == "CustomHeader"



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
import time

try:
    from unittest import mock
except ImportError:
    import mock

import pytest

from azure.core import PipelineClient
from azure.core.exceptions import ServiceResponseError
from azure.core.polling import *
from azure.core.polling.base_polling import LROBasePolling, LocationPolling

# from msrest.serialization import Model


@pytest.fixture
def client():
    # The poller itself don't use it, so we don't need something functional
    return PipelineClient("https://baseurl")


def test_abc_polling():
    abc_polling = PollingMethod()

    with pytest.raises(NotImplementedError):
        abc_polling.initialize(None, None, None)

    with pytest.raises(NotImplementedError):
        abc_polling.run()

    with pytest.raises(NotImplementedError):
        abc_polling.status()

    with pytest.raises(NotImplementedError):
        abc_polling.finished()

    with pytest.raises(NotImplementedError):
        abc_polling.resource()

    with pytest.raises(TypeError):
        abc_polling.get_continuation_token()

    with pytest.raises(TypeError):
        abc_polling.from_continuation_token("token")


def test_no_polling(client):
    no_polling = NoPolling()

    initial_response = "initial response"

    def deserialization_cb(response):
        assert response == initial_response
        return "Treated: " + response

    no_polling.initialize(client, initial_response, deserialization_cb)
    no_polling.run()  # Should no raise and do nothing
    assert no_polling.status() == "succeeded"
    assert no_polling.finished()
    assert no_polling.resource() == "Treated: " + initial_response

    continuation_token = no_polling.get_continuation_token()
    assert isinstance(continuation_token, str)

    no_polling_revived_args = NoPolling.from_continuation_token(
        continuation_token, deserialization_callback=deserialization_cb, client=client
    )
    no_polling_revived = NoPolling()
    no_polling_revived.initialize(*no_polling_revived_args)
    assert no_polling_revived.status() == "succeeded"
    assert no_polling_revived.finished()
    assert no_polling_revived.resource() == "Treated: " + initial_response


def test_polling_with_path_format_arguments(client):
    method = LROBasePolling(timeout=0, path_format_arguments={"host": "host:3000", "accountName": "local"})
    client._base_url = "http://{accountName}{host}"

    method._operation = LocationPolling()
    method._operation._location_url = "/results/1"
    method._client = client
    assert "http://localhost:3000/results/1" == method._client.format_url(
        method._operation.get_polling_url(), **method._path_format_arguments
    )


class PollingTwoSteps(PollingMethod):
    """An empty poller that returns the deserialized initial response."""

    def __init__(self, sleep=0):
        self._initial_response = None
        self._deserialization_callback = None
        self._sleep = sleep

    def initialize(self, _, initial_response, deserialization_callback):
        self._initial_response = initial_response
        self._deserialization_callback = deserialization_callback
        self._finished = False

    def run(self):
        """Empty run, no polling."""
        self._finished = True
        time.sleep(self._sleep)  # Give me time to add callbacks!

    def status(self):
        """Return the current status as a string.
        :rtype: str
        """
        return "succeeded" if self._finished else "running"

    def finished(self):
        """Is this polling finished?
        :rtype: bool
        """
        return self._finished

    def resource(self):
        return self._deserialization_callback(self._initial_response)

    def get_continuation_token(self):
        return self._initial_response

    @classmethod
    def from_continuation_token(cls, continuation_token, **kwargs):
        # type(str, Any) -> Tuple
        initial_response = continuation_token
        deserialization_callback = kwargs["deserialization_callback"]
        return None, initial_response, deserialization_callback


def test_poller(client):

    # Same the poller itself doesn't care about the initial_response, and there is no type constraint here
    initial_response = "Initial response"

    # Same for deserialization_callback, just pass to the polling_method
    def deserialization_callback(response):
        assert response == initial_response
        return "Treated: " + response

    method = NoPolling()

    poller = LROPoller(client, initial_response, deserialization_callback, method)

    done_cb = mock.MagicMock()
    poller.add_done_callback(done_cb)

    result = poller.result()
    assert poller.done()
    assert result == "Treated: " + initial_response
    assert poller.status() == "succeeded"
    assert poller.polling_method() is method
    done_cb.assert_called_once_with(method)

    # # Test with a basic Model
    # poller = LROPoller(client, initial_response, Model, method)
    # assert poller._polling_method._deserialization_callback == Model.deserialize

    # Test poller that method do a run
    method = PollingTwoSteps(sleep=1)
    poller = LROPoller(client, initial_response, deserialization_callback, method)

    done_cb = mock.MagicMock()
    done_cb2 = mock.MagicMock()
    poller.add_done_callback(done_cb)
    poller.remove_done_callback(done_cb2)

    result = poller.result()
    assert result == "Treated: " + initial_response
    assert poller.status() == "succeeded"
    done_cb.assert_called_once_with(method)
    done_cb2.assert_not_called()

    with pytest.raises(ValueError) as excinfo:
        poller.remove_done_callback(done_cb)
    assert "Process is complete" in str(excinfo.value)

    # Test continuation token
    cont_token = poller.continuation_token()

    method = PollingTwoSteps(sleep=1)
    new_poller = LROPoller.from_continuation_token(
        continuation_token=cont_token,
        client=client,
        initial_response=initial_response,
        deserialization_callback=deserialization_callback,
        polling_method=method,
    )
    result = new_poller.result()
    assert result == "Treated: " + initial_response
    assert new_poller.status() == "succeeded"


def test_broken_poller(client):
    class NoPollingError(PollingTwoSteps):
        def run(self):
            raise ValueError("Something bad happened")

    initial_response = "Initial response"

    def deserialization_callback(response):
        return "Treated: " + response

    method = NoPollingError()
    poller = LROPoller(client, initial_response, deserialization_callback, method)

    with pytest.raises(ValueError) as excinfo:
        poller.result()
    assert "Something bad happened" in str(excinfo.value)


def test_poller_error_continuation(client):
    class NoPollingError(PollingTwoSteps):
        def run(self):
            raise ServiceResponseError("Something bad happened")

    initial_response = "Initial response"

    def deserialization_callback(response):
        return "Treated: " + response

    method = NoPollingError()
    poller = LROPoller(client, initial_response, deserialization_callback, method)

    with pytest.raises(ServiceResponseError) as excinfo:
        poller.result()
    assert "Something bad happened" in str(excinfo.value)
    assert excinfo.value.continuation_token == "Initial response"



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import pytest
from azure.core.utils import case_insensitive_dict


@pytest.fixture()
def accept_cases():
    return ["accept", "Accept", "ACCEPT", "aCCePT"]


def test_case_insensitive_dict_basic(accept_cases):
    my_dict = case_insensitive_dict({"accept": "application/json"})
    for accept_case in accept_cases:
        assert my_dict[accept_case] == "application/json"


def test_case_insensitive_dict_override(accept_cases):
    for accept_case in accept_cases:
        my_dict = case_insensitive_dict({accept_case: "should-not/be-me"})
        my_dict["accept"] = "application/json"
        assert my_dict[accept_case] == my_dict["accept"] == "application/json"


def test_case_insensitive_dict_initialization():
    dict_response = {"platformUpdateDomainCount": 5, "platformFaultDomainCount": 3, "virtualMachines": []}
    a = case_insensitive_dict(platformUpdateDomainCount=5, platformFaultDomainCount=3, virtualMachines=[])
    b = case_insensitive_dict(
        zip(["platformUpdateDomainCount", "platformFaultDomainCount", "virtualMachines"], [5, 3, []])
    )
    c = case_insensitive_dict(
        [("platformFaultDomainCount", 3), ("platformUpdateDomainCount", 5), ("virtualMachines", [])]
    )
    d = case_insensitive_dict({"virtualMachines": [], "platformFaultDomainCount": 3, "platformUpdateDomainCount": 5})
    e = case_insensitive_dict({"platformFaultDomainCount": 3, "virtualMachines": []}, platformUpdateDomainCount=5)
    f = case_insensitive_dict(dict_response)
    g = case_insensitive_dict(**dict_response)
    assert a == b == c == d == e == f == g
    dicts = [a, b, c, d, e, f, g]
    for d in dicts:
        assert len(d) == 3
        assert d["platformUpdateDomainCount"] == d["platformupdatedomaincount"] == d["PLATFORMUPDATEDOMAINCOUNT"] == 5
        assert d["platformFaultDomainCount"] == d["platformfaultdomaincount"] == d["PLATFORMFAULTDOMAINCOUNT"] == 3
        assert d["virtualMachines"] == d["virtualmachines"] == d["VIRTUALMACHINES"] == []


def test_case_insensitive_dict_cant_compare():
    my_dict = case_insensitive_dict({"accept": "application/json"})
    assert my_dict != "accept"


def test_case_insensitive_dict_lowerkey_items():
    my_dict = case_insensitive_dict({"accept": "application/json"})
    assert list(my_dict.lowerkey_items()) == [("accept", "application/json")]


@pytest.mark.parametrize(
    "other, expected",
    (
        ({"PLATFORMUPDATEDOMAINCOUNT": 5}, True),
        ({}, False),
        (None, False),
    ),
)
def test_case_insensitive_dict_equality(other, expected):
    my_dict = case_insensitive_dict({"platformUpdateDomainCount": 5})
    result = my_dict == other
    assert result == expected


def test_case_insensitive_dict_keys():
    keys = ["One", "TWO", "tHrEe", "four"]
    my_dict = case_insensitive_dict({key: idx for idx, key in enumerate(keys, 1)})
    dict_keys = list(my_dict.keys())

    assert dict_keys == keys


def test_case_insensitive_copy():
    keys = ["One", "TWO", "tHrEe", "four"]
    my_dict = case_insensitive_dict({key: idx for idx, key in enumerate(keys, 1)})
    copy_my_dict = my_dict.copy()
    assert copy_my_dict is not my_dict
    assert copy_my_dict == my_dict


def test_case_insensitive_keys_present(accept_cases):
    my_dict = case_insensitive_dict({"accept": "application/json"})

    for key in accept_cases:
        assert key in my_dict


def test_case_insensitive_keys_delete(accept_cases):
    my_dict = case_insensitive_dict({"accept": "application/json"})

    for key in accept_cases:
        del my_dict[key]
        assert key not in my_dict
        my_dict[key] = "application/json"


def test_case_iter():
    keys = ["One", "TWO", "tHrEe", "four"]
    my_dict = case_insensitive_dict({key: idx for idx, key in enumerate(keys, 1)})

    for key in my_dict:
        assert key in keys



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import sys
import pytest
import json
import xml.etree.ElementTree as ET
from azure.core.pipeline.transport import HttpRequest as PipelineTransportHttpRequest
from azure.core.rest import HttpRequest as RestHttpRequest

try:
    import collections.abc as collections
except ImportError:
    import collections


@pytest.fixture
def old_request():
    return PipelineTransportHttpRequest("GET", "/")


@pytest.fixture
def new_request():
    return RestHttpRequest("GET", "/")


def test_request_attr_parity(old_request, new_request):
    for attr in dir(old_request):
        if not attr[0] == "_":
            # if not a private attr, we want parity
            assert hasattr(new_request, attr)


def test_request_set_attrs(old_request, new_request):
    for attr in dir(old_request):
        if attr[0] == "_":
            continue
        try:
            # if we can set it on the old request, we want to
            # be able to set it on the new
            setattr(old_request, attr, "foo")
        except:
            pass
        else:
            setattr(new_request, attr, "foo")
            assert getattr(old_request, attr) == getattr(new_request, attr) == "foo"


def test_request_multipart_mixed_info(old_request, new_request):
    old_request.multipart_mixed_info = "foo"
    new_request.multipart_mixed_info = "foo"
    assert old_request.multipart_mixed_info == new_request.multipart_mixed_info == "foo"


def test_request_files_attr(old_request, new_request):
    assert old_request.files == new_request.files == None
    old_request.files = {"hello": "world"}
    new_request.files = {"hello": "world"}
    assert old_request.files == new_request.files == {"hello": "world"}


def test_request_data_attr(old_request, new_request):
    assert old_request.data == new_request.data == None
    old_request.data = {"hello": "world"}
    new_request.data = {"hello": "world"}
    assert old_request.data == new_request.data == {"hello": "world"}


def test_request_query(old_request, new_request):
    assert old_request.query == new_request.query == {}
    old_request.url = "http://localhost:5000?a=b&c=d"
    new_request.url = "http://localhost:5000?a=b&c=d"
    assert old_request.query == new_request.query == {"a": "b", "c": "d"}


def test_request_query_and_params_kwarg(old_request):
    # should be same behavior if we pass in query params through the params kwarg in the new requests
    old_request.url = "http://localhost:5000?a=b&c=d"
    new_request = RestHttpRequest("GET", "http://localhost:5000", params={"a": "b", "c": "d"})
    assert old_request.query == new_request.query == {"a": "b", "c": "d"}


def test_request_body(old_request, new_request):
    assert old_request.body == new_request.body == None
    old_request.data = {"hello": "world"}
    new_request.data = {"hello": "world"}
    assert old_request.body == new_request.body == new_request.content == {"hello": "world"}
    # files will not override data
    old_request.files = {"foo": "bar"}
    new_request.files = {"foo": "bar"}
    assert old_request.body == new_request.body == new_request.content == {"hello": "world"}

    # nullify data
    old_request.data = None
    new_request.data = None
    assert old_request.data == new_request.data == old_request.body == new_request.body == None


def test_format_parameters(old_request, new_request):
    old_request.url = "a/b/c?t=y"
    new_request.url = "a/b/c?t=y"
    assert old_request.url == new_request.url == "a/b/c?t=y"
    old_request.format_parameters({"g": "h"})
    new_request.format_parameters({"g": "h"})

    # ordering can vary, so not sticking on order
    assert old_request.url in ["a/b/c?g=h&t=y", "a/b/c?t=y&g=h"]
    assert new_request.url in ["a/b/c?g=h&t=y", "a/b/c?t=y&g=h"]


def test_request_format_parameters_and_params_kwarg(old_request):
    # calling format_parameters on an old request should be the same
    # behavior as passing in params to new request
    old_request.url = "a/b/c?t=y"
    old_request.format_parameters({"g": "h"})
    new_request = RestHttpRequest("GET", "a/b/c?t=y", params={"g": "h"})
    assert old_request.url in ["a/b/c?g=h&t=y", "a/b/c?t=y&g=h"]
    assert new_request.url in ["a/b/c?g=h&t=y", "a/b/c?t=y&g=h"]

    # additionally, calling format_parameters on a new request
    # should be the same as passing the params to a new request
    assert new_request.url in ["a/b/c?g=h&t=y", "a/b/c?t=y&g=h"]
    assert new_request.url in ["a/b/c?g=h&t=y", "a/b/c?t=y&g=h"]


def test_request_streamed_data_body(old_request, new_request):
    assert old_request.files == new_request.files == None
    assert old_request.data == new_request.data == None
    old_request.files = new_request.files = "foo"
    # passing in iterable
    def streaming_body(data):
        yield data  # pragma: nocover

    old_request.set_streamed_data_body(streaming_body("i will be streamed"))
    new_request.set_streamed_data_body(streaming_body("i will be streamed"))

    assert old_request.files == new_request.files == None
    assert isinstance(old_request.data, collections.Iterable)
    assert isinstance(new_request.data, collections.Iterable)
    assert isinstance(old_request.body, collections.Iterable)
    assert isinstance(new_request.body, collections.Iterable)
    assert isinstance(new_request.content, collections.Iterable)
    assert old_request.headers == new_request.headers == {}


def test_request_streamed_data_body_non_iterable(old_request, new_request):
    # should fail before nullifying the files property
    old_request.files = new_request.files = "foo"
    # passing in non iterable
    with pytest.raises(TypeError) as ex:
        old_request.set_streamed_data_body(1)
    assert "A streamable data source must be an open file-like object or iterable" in str(ex.value)
    assert old_request.data is None
    assert old_request.files == "foo"

    with pytest.raises(TypeError) as ex:
        new_request.set_streamed_data_body(1)
    assert "A streamable data source must be an open file-like object or iterable" in str(ex.value)
    assert old_request.data is None
    assert old_request.files == "foo"
    assert old_request.headers == new_request.headers == {}


def test_request_streamed_data_body_and_content_kwarg(old_request):
    # passing stream bodies to set_streamed_data_body
    # and passing a stream body to the content kwarg of the new request should be the same
    def streaming_body(data):
        yield data  # pragma: nocover

    old_request.set_streamed_data_body(streaming_body("stream"))
    new_request = RestHttpRequest("GET", "/", content=streaming_body("stream"))
    assert old_request.files == new_request.files == None
    assert isinstance(old_request.data, collections.Iterable)
    assert isinstance(new_request.data, collections.Iterable)
    assert isinstance(old_request.body, collections.Iterable)
    assert isinstance(new_request.body, collections.Iterable)
    assert isinstance(new_request.content, collections.Iterable)
    assert old_request.headers == new_request.headers == {}


def test_request_text_body(old_request, new_request):
    assert old_request.files == new_request.files == None
    assert old_request.data == new_request.data == None
    old_request.files = new_request.files = "foo"
    old_request.set_text_body("i am text")
    new_request.set_text_body("i am text")

    assert old_request.files == new_request.files == None
    assert (
        old_request.data
        == new_request.data
        == old_request.body
        == new_request.body
        == new_request.content
        == "i am text"
    )
    assert old_request.headers["Content-Length"] == new_request.headers["Content-Length"] == "9"
    assert not old_request.headers.get("Content-Type")
    assert new_request.headers["Content-Type"] == "text/plain"


def test_request_text_body_and_content_kwarg(old_request):
    old_request.set_text_body("i am text")
    new_request = RestHttpRequest("GET", "/", content="i am text")
    assert (
        old_request.data
        == new_request.data
        == old_request.body
        == new_request.body
        == new_request.content
        == "i am text"
    )
    assert old_request.headers["Content-Length"] == new_request.headers["Content-Length"] == "9"
    assert old_request.files == new_request.files == None


def test_request_xml_body(old_request, new_request):
    assert old_request.files == new_request.files == None
    assert old_request.data == new_request.data == None
    old_request.files = new_request.files = "foo"
    xml_input = ET.Element("root")
    old_request.set_xml_body(xml_input)
    new_request.set_xml_body(xml_input)

    assert old_request.files == new_request.files == None
    assert (
        old_request.data
        == new_request.data
        == old_request.body
        == new_request.body
        == new_request.content
        == b"<?xml version='1.0' encoding='utf-8'?>\n<root />"
    )
    assert old_request.headers == new_request.headers == {"Content-Length": "47"}


def test_request_xml_body_and_content_kwarg(old_request):
    old_request.set_text_body("i am text")
    new_request = RestHttpRequest("GET", "/", content="i am text")
    assert (
        old_request.data
        == new_request.data
        == old_request.body
        == new_request.body
        == new_request.content
        == "i am text"
    )
    assert old_request.headers["Content-Length"] == new_request.headers["Content-Length"] == "9"
    assert old_request.files == new_request.files == None


def test_request_json_body(old_request, new_request):
    assert old_request.files == new_request.files == None
    assert old_request.data == new_request.data == None
    old_request.files = new_request.files = "foo"
    json_input = {"hello": "world"}
    old_request.set_json_body(json_input)
    new_request.set_json_body(json_input)

    assert old_request.files == new_request.files == None
    assert (
        old_request.data
        == new_request.data
        == old_request.body
        == new_request.body
        == new_request.content
        == json.dumps(json_input)
    )
    assert old_request.headers["Content-Length"] == new_request.headers["Content-Length"] == "18"
    assert not old_request.headers.get("Content-Type")
    assert new_request.headers["Content-Type"] == "application/json"


def test_request_json_body_and_json_kwarg(old_request):
    json_input = {"hello": "world"}
    old_request.set_json_body(json_input)
    new_request = RestHttpRequest("GET", "/", json=json_input)
    assert (
        old_request.data
        == new_request.data
        == old_request.body
        == new_request.body
        == new_request.content
        == json.dumps(json_input)
    )
    assert old_request.headers["Content-Length"] == new_request.headers["Content-Length"] == "18"
    assert not old_request.headers.get("Content-Type")
    assert new_request.headers["Content-Type"] == "application/json"
    assert old_request.files == new_request.files == None


def test_request_formdata_body_files(old_request, new_request):
    assert old_request.files == new_request.files == None
    assert old_request.data == new_request.data == None
    old_request.data = new_request.data = "foo"
    old_request.files = new_request.files = "bar"

    # without the urlencoded content type, set_formdata_body
    # will set it as files
    old_request.set_formdata_body({"fileName": "hello.jpg"})
    new_request.set_formdata_body({"fileName": "hello.jpg"})

    assert old_request.data == new_request.data == None
    assert old_request.files == new_request.files == new_request.content == {"fileName": (None, "hello.jpg")}

    # we don't set any multipart headers with boundaries
    # we rely on the transport to boundary calculating
    assert old_request.headers == new_request.headers == {}


def test_request_formdata_body_data(old_request, new_request):
    assert old_request.files == new_request.files == None
    assert old_request.data == new_request.data == None
    old_request.data = new_request.data = "foo"
    old_request.files = new_request.files = "bar"

    # with the urlencoded content type, set_formdata_body
    # will set it as data
    old_request.headers["Content-Type"] = "application/x-www-form-urlencoded"
    new_request.headers["Content-Type"] = "application/x-www-form-urlencoded"
    old_request.set_formdata_body({"fileName": "hello.jpg"})
    new_request.set_formdata_body({"fileName": "hello.jpg"})

    assert old_request.files == new_request.files == None
    assert (
        old_request.data
        == new_request.data
        == old_request.body
        == new_request.body
        == new_request.content
        == {"fileName": "hello.jpg"}
    )
    # old behavior would pop out the Content-Type header
    # new behavior doesn't do that
    assert old_request.headers == {}
    assert new_request.headers == {"Content-Type": "application/x-www-form-urlencoded"}


def test_request_formdata_body_and_files_kwarg(old_request):
    files = {"fileName": "hello.jpg"}
    old_request.set_formdata_body(files)
    new_request = RestHttpRequest("GET", "/", files=files)
    assert old_request.data == new_request.data == None
    assert old_request.body == new_request.body == None
    assert old_request.headers == new_request.headers == {}
    assert old_request.files == new_request.files == {"fileName": (None, "hello.jpg")}


def test_request_formdata_body_and_data_kwarg(old_request):
    data = {"fileName": "hello.jpg"}
    # with the urlencoded content type, set_formdata_body
    # will set it as data
    old_request.headers["Content-Type"] = "application/x-www-form-urlencoded"
    old_request.set_formdata_body(data)
    new_request = RestHttpRequest("GET", "/", data=data)
    assert (
        old_request.data
        == new_request.data
        == old_request.body
        == new_request.body
        == new_request.content
        == {"fileName": "hello.jpg"}
    )
    assert old_request.headers == {}
    assert new_request.headers == {"Content-Type": "application/x-www-form-urlencoded"}
    assert old_request.files == new_request.files == None


def test_request_bytes_body(old_request, new_request):
    assert old_request.files == new_request.files == None
    assert old_request.data == new_request.data == None
    old_request.files = new_request.files = "foo"
    bytes_input = b"hello, world!"
    old_request.set_bytes_body(bytes_input)
    new_request.set_bytes_body(bytes_input)

    assert old_request.files == new_request.files == None
    assert (
        old_request.data
        == new_request.data
        == old_request.body
        == new_request.body
        == new_request.content
        == bytes_input
    )
    assert old_request.headers == new_request.headers == {"Content-Length": "13"}


def test_request_bytes_body_and_content_kwarg(old_request):
    bytes_input = b"hello, world!"
    old_request.set_bytes_body(bytes_input)
    new_request = RestHttpRequest("GET", "/", content=bytes_input)
    assert (
        old_request.data
        == new_request.data
        == old_request.body
        == new_request.body
        == new_request.content
        == bytes_input
    )
    assert old_request.headers == new_request.headers == {"Content-Length": "13"}
    assert old_request.files == new_request.files



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
import os
import pytest
from azure.core import AsyncPipelineClient
from azure.core.exceptions import DecodeError
from utils import HTTP_REQUESTS


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_decompress_plain_no_header(http_request):
    # expect plain text
    account_name = "coretests"
    account_url = "https://{}.blob.core.windows.net".format(account_name)
    url = "https://{}.blob.core.windows.net/tests/test.txt".format(account_name)
    client = AsyncPipelineClient(account_url)
    async with client:
        request = http_request("GET", url)
        pipeline_response = await client._pipeline.run(request, stream=True)
        response = pipeline_response.http_response
        data = response.stream_download(client._pipeline, decompress=True)
        content = b""
        async for d in data:
            content += d
        decoded = content.decode("utf-8")
        assert decoded == "test"


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_compress_plain_no_header(http_request):
    # expect plain text
    account_name = "coretests"
    account_url = "https://{}.blob.core.windows.net".format(account_name)
    url = "https://{}.blob.core.windows.net/tests/test.txt".format(account_name)
    client = AsyncPipelineClient(account_url)
    async with client:
        request = http_request("GET", url)
        pipeline_response = await client._pipeline.run(request, stream=True)
        response = pipeline_response.http_response
        data = response.stream_download(client._pipeline, decompress=False)
        content = b""
        async for d in data:
            content += d
        decoded = content.decode("utf-8")
        assert decoded == "test"


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_decompress_compressed_no_header(http_request):
    # expect compressed text
    account_name = "coretests"
    account_url = "https://{}.blob.core.windows.net".format(account_name)
    url = "https://{}.blob.core.windows.net/tests/test.tar.gz".format(account_name)
    client = AsyncPipelineClient(account_url)
    async with client:
        request = http_request("GET", url)
        pipeline_response = await client._pipeline.run(request, stream=True)
        response = pipeline_response.http_response
        data = response.stream_download(client._pipeline, decompress=True)
        content = b""
        async for d in data:
            content += d
        try:
            decoded = content.decode("utf-8")
            assert False
        except UnicodeDecodeError:
            pass


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_compress_compressed_no_header(http_request):
    # expect compressed text
    account_name = "coretests"
    account_url = "https://{}.blob.core.windows.net".format(account_name)
    url = "https://{}.blob.core.windows.net/tests/test.tar.gz".format(account_name)
    client = AsyncPipelineClient(account_url)
    async with client:
        request = http_request("GET", url)
        pipeline_response = await client._pipeline.run(request, stream=True)
        response = pipeline_response.http_response
        data = response.stream_download(client._pipeline, decompress=False)
        content = b""
        async for d in data:
            content += d
        try:
            decoded = content.decode("utf-8")
            assert False
        except UnicodeDecodeError:
            pass


@pytest.mark.live_test_only
@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_decompress_plain_header(http_request):
    # expect error
    import zlib

    account_name = "coretests"
    account_url = "https://{}.blob.core.windows.net".format(account_name)
    url = "https://{}.blob.core.windows.net/tests/test_with_header.txt".format(account_name)
    client = AsyncPipelineClient(account_url)
    async with client:
        request = http_request("GET", url)
        pipeline_response = await client._pipeline.run(request, stream=True)
        response = pipeline_response.http_response
        data = response.stream_download(client._pipeline, decompress=True)
        try:
            content = b""
            async for d in data:
                content += d
            assert False
        except (zlib.error, DecodeError):
            pass


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_compress_plain_header(http_request):
    # expect plain text
    account_name = "coretests"
    account_url = "https://{}.blob.core.windows.net".format(account_name)
    url = "https://{}.blob.core.windows.net/tests/test_with_header.txt".format(account_name)
    client = AsyncPipelineClient(account_url)
    async with client:
        request = http_request("GET", url)
        pipeline_response = await client._pipeline.run(request, stream=True)
        response = pipeline_response.http_response
        data = response.stream_download(client._pipeline, decompress=False)
        content = b""
        async for d in data:
            content += d
        decoded = content.decode("utf-8")
        assert decoded == "test"


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_decompress_compressed_header(http_request):
    # expect plain text
    account_name = "coretests"
    account_url = "https://{}.blob.core.windows.net".format(account_name)
    url = "https://{}.blob.core.windows.net/tests/test_with_header.tar.gz".format(account_name)
    client = AsyncPipelineClient(account_url)
    async with client:
        request = http_request("GET", url)
        pipeline_response = await client._pipeline.run(request, stream=True)
        response = pipeline_response.http_response
        data = response.stream_download(client._pipeline, decompress=True)
        content = b""
        async for d in data:
            content += d
        decoded = content.decode("utf-8")
        assert decoded == "test"


@pytest.mark.live_test_only
@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_compress_compressed_header(http_request):
    # expect compressed text
    account_name = "coretests"
    account_url = "https://{}.blob.core.windows.net".format(account_name)
    url = "https://{}.blob.core.windows.net/tests/test_with_header.tar.gz".format(account_name)
    client = AsyncPipelineClient(account_url)
    async with client:
        request = http_request("GET", url)
        pipeline_response = await client._pipeline.run(request, stream=True)
        response = pipeline_response.http_response
        data = response.stream_download(client._pipeline, decompress=False)
        content = b""
        async for d in data:
            content += d
        try:
            decoded = content.decode("utf-8")
            assert False
        except UnicodeDecodeError:
            pass



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
"""Tests for the HttpLoggingPolicy."""
import pytest
import logging
import types

try:
    from unittest.mock import Mock
except ImportError:  # python < 3.3
    from mock import Mock  # type: ignore
from azure.core.pipeline import PipelineResponse, PipelineRequest, PipelineContext
from azure.core.pipeline.policies import (
    HttpLoggingPolicy,
)
from utils import HTTP_RESPONSES, create_http_response, request_and_responses_product
from azure.core.pipeline._tools import is_rest


@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(HTTP_RESPONSES))
def test_http_logger(http_request, http_response):
    class MockHandler(logging.Handler):
        def __init__(self):
            super(MockHandler, self).__init__()
            self.messages = []

        def reset(self):
            self.messages = []

        def emit(self, record):
            self.messages.append(record)

    mock_handler = MockHandler()

    logger = logging.getLogger("testlogger")
    logger.addHandler(mock_handler)
    logger.setLevel(logging.DEBUG)

    policy = HttpLoggingPolicy(logger=logger)

    universal_request = http_request("GET", "http://localhost/")
    http_response = create_http_response(http_response, universal_request, None)
    http_response.status_code = 202
    http_response.headers["x-ms-error-code"] = "ERRORCODE"
    request = PipelineRequest(universal_request, PipelineContext(None))

    # Basics

    policy.on_request(request)
    response = PipelineResponse(request, http_response, request.context)
    policy.on_response(request, response)

    assert all(m.levelname == "INFO" for m in mock_handler.messages)
    assert len(mock_handler.messages) == 2
    messages_request = mock_handler.messages[0].message.split("\n")
    messages_response = mock_handler.messages[1].message.split("\n")
    assert messages_request[0] == "Request URL: 'http://localhost/'"
    assert messages_request[1] == "Request method: 'GET'"
    assert messages_request[2] == "Request headers:"
    assert messages_request[3] == "No body was attached to the request"
    assert messages_response[0] == "Response status: 202"
    assert messages_response[1] == "Response headers:"
    assert messages_response[2] == "    'x-ms-error-code': 'ERRORCODE'"

    mock_handler.reset()

    # Let's make this request a failure, retried twice

    policy.on_request(request)
    response = PipelineResponse(request, http_response, request.context)
    policy.on_response(request, response)

    policy.on_request(request)
    response = PipelineResponse(request, http_response, request.context)
    policy.on_response(request, response)

    assert all(m.levelname == "INFO" for m in mock_handler.messages)
    assert len(mock_handler.messages) == 4
    messages_request1 = mock_handler.messages[0].message.split("\n")
    messages_response1 = mock_handler.messages[1].message.split("\n")
    messages_request2 = mock_handler.messages[2].message.split("\n")
    messages_response2 = mock_handler.messages[3].message.split("\n")
    assert messages_request1[0] == "Request URL: 'http://localhost/'"
    assert messages_request1[1] == "Request method: 'GET'"
    assert messages_request1[2] == "Request headers:"
    assert messages_request1[3] == "No body was attached to the request"
    assert messages_response1[0] == "Response status: 202"
    assert messages_response1[1] == "Response headers:"
    assert messages_request2[0] == "Request URL: 'http://localhost/'"
    assert messages_request2[1] == "Request method: 'GET'"
    assert messages_request2[2] == "Request headers:"
    assert messages_request2[3] == "No body was attached to the request"
    assert messages_response2[0] == "Response status: 202"
    assert messages_response2[1] == "Response headers:"

    mock_handler.reset()

    # Headers and query parameters

    policy.allowed_query_params = ["country"]

    universal_request.headers = {
        "Accept": "Caramel",
        "Hate": "Chocolat",
    }
    http_response.headers = {
        "Content-Type": "Caramel",
        "HateToo": "Chocolat",
    }
    universal_request.url = "http://localhost/?country=france&city=aix"

    policy.on_request(request)
    response = PipelineResponse(request, http_response, request.context)
    policy.on_response(request, response)

    assert all(m.levelname == "INFO" for m in mock_handler.messages)
    assert len(mock_handler.messages) == 2
    messages_request = mock_handler.messages[0].message.split("\n")
    messages_response = mock_handler.messages[1].message.split("\n")
    assert messages_request[0] == "Request URL: 'http://localhost/?country=france&city=REDACTED'"
    assert messages_request[1] == "Request method: 'GET'"
    assert messages_request[2] == "Request headers:"
    # Dict not ordered in Python, exact logging order doesn't matter
    assert set([messages_request[3], messages_request[4]]) == set(["    'Accept': 'Caramel'", "    'Hate': 'REDACTED'"])
    assert messages_request[5] == "No body was attached to the request"
    assert messages_response[0] == "Response status: 202"
    assert messages_response[1] == "Response headers:"
    # Dict not ordered in Python, exact logging order doesn't matter
    assert set([messages_response[2], messages_response[3]]) == set(
        ["    'Content-Type': 'Caramel'", "    'HateToo': 'REDACTED'"]
    )

    mock_handler.reset()


@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(HTTP_RESPONSES))
def test_http_logger_operation_level(http_request, http_response):
    class MockHandler(logging.Handler):
        def __init__(self):
            super(MockHandler, self).__init__()
            self.messages = []

        def reset(self):
            self.messages = []

        def emit(self, record):
            self.messages.append(record)

    mock_handler = MockHandler()

    logger = logging.getLogger("testlogger")
    logger.addHandler(mock_handler)
    logger.setLevel(logging.DEBUG)

    policy = HttpLoggingPolicy()
    kwargs = {"logger": logger}

    universal_request = http_request("GET", "http://localhost/")
    http_response = create_http_response(http_response, universal_request, None)
    http_response.status_code = 202
    request = PipelineRequest(universal_request, PipelineContext(None, **kwargs))

    # Basics

    policy.on_request(request)
    response = PipelineResponse(request, http_response, request.context)
    policy.on_response(request, response)

    assert all(m.levelname == "INFO" for m in mock_handler.messages)
    assert len(mock_handler.messages) == 2
    messages_request = mock_handler.messages[0].message.split("\n")
    messages_response = mock_handler.messages[1].message.split("\n")
    assert messages_request[0] == "Request URL: 'http://localhost/'"
    assert messages_request[1] == "Request method: 'GET'"
    assert messages_request[2] == "Request headers:"
    assert messages_request[3] == "No body was attached to the request"
    assert messages_response[0] == "Response status: 202"
    assert messages_response[1] == "Response headers:"

    mock_handler.reset()

    # Let's make this request a failure, retried twice

    request = PipelineRequest(universal_request, PipelineContext(None, **kwargs))

    policy.on_request(request)
    response = PipelineResponse(request, http_response, request.context)
    policy.on_response(request, response)

    policy.on_request(request)
    response = PipelineResponse(request, http_response, request.context)
    policy.on_response(request, response)

    assert all(m.levelname == "INFO" for m in mock_handler.messages)
    assert len(mock_handler.messages) == 4
    messages_request1 = mock_handler.messages[0].message.split("\n")
    messages_response1 = mock_handler.messages[1].message.split("\n")
    messages_request2 = mock_handler.messages[2].message.split("\n")
    messages_response2 = mock_handler.messages[3].message.split("\n")
    assert messages_request1[0] == "Request URL: 'http://localhost/'"
    assert messages_request1[1] == "Request method: 'GET'"
    assert messages_request1[2] == "Request headers:"
    assert messages_request1[3] == "No body was attached to the request"
    assert messages_response1[0] == "Response status: 202"
    assert messages_response1[1] == "Response headers:"
    assert messages_request2[0] == "Request URL: 'http://localhost/'"
    assert messages_request2[1] == "Request method: 'GET'"
    assert messages_request2[2] == "Request headers:"
    assert messages_request2[3] == "No body was attached to the request"
    assert messages_response2[0] == "Response status: 202"
    assert messages_response2[1] == "Response headers:"

    mock_handler.reset()


@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(HTTP_RESPONSES))
def test_http_logger_with_body(http_request, http_response):
    class MockHandler(logging.Handler):
        def __init__(self):
            super(MockHandler, self).__init__()
            self.messages = []

        def reset(self):
            self.messages = []

        def emit(self, record):
            self.messages.append(record)

    mock_handler = MockHandler()

    logger = logging.getLogger("testlogger")
    logger.addHandler(mock_handler)
    logger.setLevel(logging.DEBUG)

    policy = HttpLoggingPolicy(logger=logger)

    universal_request = http_request("GET", "http://localhost/")
    universal_request.body = "testbody"
    http_response = create_http_response(http_response, universal_request, None)
    http_response.status_code = 202
    request = PipelineRequest(universal_request, PipelineContext(None))

    policy.on_request(request)
    response = PipelineResponse(request, http_response, request.context)
    policy.on_response(request, response)

    assert all(m.levelname == "INFO" for m in mock_handler.messages)
    assert len(mock_handler.messages) == 2
    messages_request = mock_handler.messages[0].message.split("\n")
    messages_response = mock_handler.messages[1].message.split("\n")
    assert messages_request[0] == "Request URL: 'http://localhost/'"
    assert messages_request[1] == "Request method: 'GET'"
    assert messages_request[2] == "Request headers:"
    assert messages_request[3] == "A body is sent with the request"
    assert messages_response[0] == "Response status: 202"
    assert messages_response[1] == "Response headers:"

    mock_handler.reset()


@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(HTTP_RESPONSES))
def test_http_logger_with_generator_body(http_request, http_response):
    class MockHandler(logging.Handler):
        def __init__(self):
            super(MockHandler, self).__init__()
            self.messages = []

        def reset(self):
            self.messages = []

        def emit(self, record):
            self.messages.append(record)

    mock_handler = MockHandler()

    logger = logging.getLogger("testlogger")
    logger.addHandler(mock_handler)
    logger.setLevel(logging.DEBUG)

    policy = HttpLoggingPolicy(logger=logger)

    universal_request = http_request("GET", "http://localhost/")
    mock = Mock()
    mock.__class__ = types.GeneratorType
    universal_request.body = mock
    http_response = create_http_response(http_response, universal_request, None)
    http_response.status_code = 202
    request = PipelineRequest(universal_request, PipelineContext(None))

    policy.on_request(request)
    response = PipelineResponse(request, http_response, request.context)
    policy.on_response(request, response)

    assert all(m.levelname == "INFO" for m in mock_handler.messages)
    assert len(mock_handler.messages) == 2
    messages_request = mock_handler.messages[0].message.split("\n")
    messages_response = mock_handler.messages[1].message.split("\n")
    assert messages_request[0] == "Request URL: 'http://localhost/'"
    assert messages_request[1] == "Request method: 'GET'"
    assert messages_request[2] == "Request headers:"
    assert messages_request[3] == "File upload"
    assert messages_response[0] == "Response status: 202"
    assert messages_response[1] == "Response headers:"

    mock_handler.reset()



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
from enum import Enum

from azure.core import CaseInsensitiveEnumMeta


class MyCustomEnum(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    FOO = "foo"
    BAR = "bar"


def test_case_insensitive_enums():
    assert MyCustomEnum.foo.value == "foo"
    assert MyCustomEnum.FOO.value == "foo"
    assert MyCustomEnum("bar").value == "bar"
    assert "bar" == MyCustomEnum.BAR
    assert "bar" == MyCustomEnum.bar
    assert MyCustomEnum["foo"] == "foo"
    assert MyCustomEnum["FOO"] == "foo"
    assert isinstance(MyCustomEnum.BAR, str)



# coding: utf-8
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------

from flask import Flask, Response
import os
from .test_routes import (
    basic_api,
    encoding_api,
    errors_api,
    polling_api,
    streams_api,
    urlencoded_api,
    multipart_api,
    xml_api,
    headers_api,
)

app = Flask(__name__)
app.register_blueprint(basic_api, url_prefix="/basic")
app.register_blueprint(encoding_api, url_prefix="/encoding")
app.register_blueprint(errors_api, url_prefix="/errors")
app.register_blueprint(polling_api, url_prefix="/polling")
app.register_blueprint(streams_api, url_prefix="/streams")
app.register_blueprint(urlencoded_api, url_prefix="/urlencoded")
app.register_blueprint(multipart_api, url_prefix="/multipart")
app.register_blueprint(xml_api, url_prefix="/xml")
app.register_blueprint(headers_api, url_prefix="/headers")


@app.route("/health", methods=["GET"])
def latin_1_charset_utf8():
    return Response(status=200)


if __name__ == "__main__":
    debugArg = True
    if os.getenv("TF_BUILD", None):
        debugArg = False
    app.run(debug=debugArg)
